import os
#from shutil import copyfile
from pathlib import Path

from .ann_convertors import check_yolov5_yaml, sa,sb
# sa, sb = f'{os.sep}images{os.sep}', f'{os.sep}labels{os.sep}'# /images/, /labels/ substrings
from .ann_convertors import yolov5_img_video_formats, \
    sa, sb, get_ann_format, save_yolo_ann, get_yolov5_ann, ann_ext_dict
from .TerraYolo import save_data_yaml
from .general import fast_scandir, run_fast_scandir



def get_yolo_lab_dir(img_dir):
    return sb.join((str(Path(img_dir)) + os.sep).rsplit(sa, 1))


def get_img_files(img_dir, format_set=frozenset(yolov5_img_video_formats)):
    return [f_name for f_name in os.listdir(img_dir) if
            os.path.splitext(f_name)[1] in format_set]


def explore_img_dir(base_dir_or_list):
    # этим методом возвращаем все словарь, ключи - подкаталоги с изображениями
    # значения число изображений
    if isinstance(base_dir_or_list, (list)):
        subfolders = base_dir_or_list
    else:
        subfolders = fast_scandir(base_dir_or_list)
    result = dict()
    if subfolders:
        for subdir in subfolders:
            subdir_img_files = get_img_files(subdir)
            if subdir_img_files:
                result[subdir] = len(subdir_img_files)
    return result


def check_yolov5_txt_for_img(img_dir, lab_dir, verbose=False):
    if not os.path.isdir(img_dir):
        if verbose:
            print(f'Каталог изображений {img_dir} не найден')
        return ''

    img_f_names = set([os.path.splitext(f_name)[0]
                       for f_name in os.listdir(img_dir)
                       if os.path.splitext(f_name)[1] in yolov5_img_video_formats])
    if len(img_f_names) < 1:
        if verbose:
            print(f'В каталоге {img_dir} не найдены файлы с изображениями')
        return False

    # чтобы получился путь к меткам, и последний каталог должен быть строго /images/
    # просто /images без слэша в конце будет ошибка. Если не последний - все равно,

    if not os.path.isdir(lab_dir):
        if verbose:
            print(f'Каталог меток {lab_dir} не найден')
        return False

    lab_f_names = set([os.path.splitext(f_name)[0]
                       for f_name in os.listdir(lab_dir)
                       if os.path.splitext(f_name)[1] == '.txt'])

    if len(img_f_names & lab_f_names) < 1:
        if verbose:
            print(f'В каталоге {img_dir} не найдены подходящие аннотации')
        return False

    return True


def check_yolov5_img_dir_name(img_dir):
    if os.path.normpath(img_dir) == os.path.normpath(get_yolo_lab_dir(img_dir)):
        return False
    return True


def img_dir_name_analyze(img_dir_dict, f_name_prior=True):
    # сначала проверяем, что потенциальное имя каталога с изображением годится для yolov5
    # строим список каталогов с метками и если совпадают, то не годится
    lab_dir_list = []
    for img_dir in img_dir_dict.keys():
        lab_dir_list.append(get_yolo_lab_dir(str(img_dir) + os.sep))
        '''           
        if os.path.isdir(lab_dir):
            lab_files = [f_name for f_name in lab_dir if os.path.splitext(f_name)=='.txt']
            lab_dir_dict[lab_dir]=len(lab_files)
        else:
            lab_dir_dict[lab_dir]=0
        '''
    bad_name_img_dir = [img_dir for ind, img_dir in enumerate(img_dir_dict.keys()) if
                        img_dir == lab_dir_list[ind]]
    for img_dir in bad_name_img_dir:
        print(
            f'Каталог с изображениями {img_dir} не содержит в пути стандартной папки {sa}')
        # delete img_dir_dict[img_dir]

    img_train_dir, img_val_dir, img_test_dir = '', '', ''
    if len(img_dir_dict) < 1:
        print('Не найдены каталоги с изображениями')
        return img_train_dir, img_val_dir, img_test_dir

    # ищем среди имен каталогов train, test, val(valid)
    train_dict = dict()
    val_dict = dict()
    test_dict = dict()

    if f_name_prior == True:
        if len(img_dir_dict) > 1:
            for img_dir in img_dir_dict.keys():
                path_parts_reversed = Path(img_dir).resolve().parts[::-1]
                # если среди каталогов есть 'train'
                if 'train' in path_parts_reversed:
                    # добавляем в соотвествующий словарь макс уровень на котором нашли train
                    # максимальный на случай train может встретиться в пути и не один раз
                    # train и другие ключи типа val тоже могут встретиться в пути
                    train_dict[img_dir] = len(path_parts_reversed
                                              ) - path_parts_reversed.index('train')

                # val
                val_pos = -1
                if 'val' in path_parts_reversed:
                    val_pos = len(path_parts_reversed
                                  ) - path_parts_reversed.index('val')

                if 'valid' in path_parts_reversed:
                    # добавляем в соотвествующий словарь максимальный уровень
                    val_pos = max(len(path_parts_reversed
                                      ) - path_parts_reversed.index('valid'), val_pos)

                if val_pos >= 0:
                    val_dict[img_dir] = val_pos
                # test

                if 'test' in path_parts_reversed:
                    # добавляем в соотвествующий словарь максимальный уровень
                    test_dict[img_dir] = len(path_parts_reversed
                                             ) - path_parts_reversed.index('test')

            def get_dir_from_max_level(dir_dict):
                max_dir_level = max(dir_dict.values())
                max_dir_ind = list(dir_dict.values()).index(max_dir_level)
                img_dir = list(dir_dict.keys())[max_dir_ind]
                return img_dir

            # если f_name_prior==True и в каталогах с изображениями встретился каталог 'train'
            if train_dict:
                img_train_dir = get_dir_from_max_level(train_dict)

                if img_train_dir in val_dict.keys():
                    del val_dict[img_train_dir]
                if img_train_dir in test_dict.keys():
                    del test_dict[img_train_dir]

            # если train не найден, то остальное ('val', 'test') игнорируем
            if val_dict:
                img_val_dir = get_dir_from_max_level(val_dict)
                if img_val_dir in test_dict.keys():
                    del test_dict[yaml_dict['val']]

            if test_dict:
                img_test_dir = get_dir_from_max_level(test_dict)

    sorted_img_dir_dict = dict(sorted(img_dir_dict.items(), key=lambda item: item[1]))

    # если не находим по названиям - считаем картинки и в img_train_dir назначаем
    # каталог с max количеством изображений
    if not img_train_dir:
        img_train_dir = list(sorted_img_dir_dict.keys())[0]
        del sorted_img_dir_dict[img_train_dir]

    if not img_val_dir:
        if sorted_img_dir_dict:
            img_val_dir = list(sorted_img_dir_dict.keys())[0]
            del sorted_img_dir_dict[img_val_dir]

    # в случае подсчета по количеству изображений test не ищем
    # if not img_test_dir:

    if (img_test_dir != '') & (img_val_dir == ''):
        img_val_dir = img_test_dir

    if (img_val_dir != '') & (img_train_dir == ''):
        img_train_dir = img_val_dir

    if (img_train_dir != '') & (img_val_dir == ''):
        img_val_dir = img_train_dir

    return img_train_dir, img_val_dir, img_test_dir

### общие соображения
# класс привязан к структуре каталогов и содержит от одного подраздела подкатлога
# yaml для каталогов имеет вспомогательное значение, для классов - основное
# смысл в том, что класс для того, чтобы получить yaml,
# если yaml есть - то и класс не так чтоб нужен
# ну еще как развитие - хранить результаты в классе, но это не точно
# train - val - test - все остальное misc, в общем случае
# привязываемся к именам - train , val-valid, test
# нет train с аннотациями - нет датасета
# потому если нет имени train - выбираем для train самый большой каталог
# в test может не быть аннотаций
# у val-valid должны быть аннотации, потому, если нет имени val-valid
# но есть test и для test есть аннотации - test становиться val-valid
# если testa нет или нет для него аннотаций - train становится val
#


class YoloV5Dataset:
    def __init__(self,
                 yaml_path_dir=None,
                 train_img_subdir='', val_img_subdir='', test_img_subdir='',
                 # train_lab_dir='', val_lab_dir='', test_lab_dir='',
                 # data_root_path=None,
                 # cl_labels=None,
                 names=(),
                 new_yaml_path=None, new_yaml_f_name='dataset.yaml',
                 f_autosave=False, f_overwrite=False,
                 ):

        if not os.path.exists(yaml_path_dir):
            print(f'файл/каталог {yaml_path_dir} с данными не найден')
            return None

        self.data_keys = ['train', 'val', 'test']
        self.dest_key = 'yolov5'
        self.data_dict = dict()
        self.yaml_dict = dict()

        self.new_yaml_path = new_yaml_path
        self.new_yaml_f_name = new_yaml_f_name
        self.f_autosave = f_autosave
        self.f_overwrite = f_overwrite

        for k in self.data_keys:
            self.data_dict[k] = dict()

        # на этом шаге ищем yaml описания
        if os.path.isfile(yaml_path_dir):
            self.base_dir = os.path.split(yaml_path_dir)
            self.parse_yaml(yaml_path=yaml_path_dir)
        elif os.path.isdir(yaml_path_dir):
            self.base_dir = yaml_path_dir
            # прочие yaml будем искать только при отсутствии списка классов
            if len(names) < 1:
                self.explore_base_dir_yaml()
            else:
                print('дан список классов, пропускаем проверку yaml - файлов')

        # если дошли до сюда и yaml_dict значит найден подходящий yaml можно уходить
        if self.yaml_dict:
            return

        print(
            f'По адресу {yaml_path_dir} не найдены подходящие yaml-описания датасетов, продолжаем анализ')

        if train_img_subdir:
            train_img_dir = os.path.join(self.base_dir, train_img_subdir)
            if not check_yolov5_img_dir_name(train_img_dir):
                print(
                    f'Имя {train_img_dir} должно содержать {sa} и не может быть каталогом изображений yolov5')
                return None
            if val_img_subdir:
                val_img_dir = os.path.join(self.base_dir, val_img_subdir)
            if test_img_subdir:
                test_img_dir = os.path.join(self.base_dir, test_img_subdir)
        else:
            train_img_dir, val_img_subdir = None, None

        if len(names) > 0:
            self.names = names
            self.nc = len(self.names)
            if not train_img_dir:
                # в этом случае будут проверены имена подкаталогов
                train_img_dir, val_img_dir, test_img_dir = self.get_img_dir_list()
                if self.yaml_dict:
                    return
            for img_dir in [train_img_dir, val_img_dir]:
                if not check_yolov5_txt_for_img(img_dir, get_yolo_lab_dir(img_dir)):
                    del self.names
                    del self.nc
                    return None
                tmp_dict = dict()
                tmp_dict['train'] = train_img_dir + os.sep
                tmp_dict['val'] = val_img_dir + os.sep
                if test_img_dir: tmp_dict['test'] = test_img_dir + os.sep
                tmp_dict['path'] = self.base_dir
                tmp_dict['names'] = self.names
                tmp_dict['nc'] = self.nc
                # yaml_path
                self.set_class_param(tmp_dict, self.base_dir, yaml_path=None)
                # self.save_yaml()
                return

        # len(names)==0,
        img_dict = explore_img_dir(self.base_dir)
        train_img_dir, val_img_dir, test_img_dir = img_dir_name_analyze(img_dict,
                                                                        f_name_prior=True)
        if not train_img_dir:
            print(f'В каталоге {self.base_dir} не найдены каталоги с изображениями')
            return None
        # если дошли до сюда, то остается искать только аннотации других форматов

        format_founded = get_ann_format(self.base_dir)
        # format_founded = set(format_founded)-set([self.dest_key])
        if len(format_founded) > 0:
            print(f'Найдены аннотации следующих форматов: {format_founded}')
            for f_key in format_founded:
                if f_key != self.dest_key:
                    # остальные форматы не рассматриваем, берем только первый подходящий
                    if self.get_labels(train_img_dir, val_img_dir, test_img_dir, f_key):
                        return
        else:
            print(f'в каталоге {yaml_path_dir} не найдены подходящие аннотации')
        return None

    def parse_yaml(self, yaml_path):
        print('parse_yaml', yaml_path)
        yaml_dict, base_dir = check_yolov5_yaml(yaml_path)
        print('yaml_dict, base_dir', base_dir)
        if yaml_dict:
            print('parse_yaml set_class_param')
            self.set_class_param(yaml_dict, self.base_dir, yaml_path)

    def get_img_dir_list(self):
        # если заданы каталоги train_img_dir, val_img_dir, test_img_dir они будут в приоритете
        # возможны ситуации нет каталога - нет изображений -нет меток-

        # img_lab_dir_dict = dict(zip([train_img_dir, val_img_dir, test_img_dir],
        # [train_lab_dir, val_lab_dir, test_lab_dir]))

        def check_tmp_yaml():
            def get_key_dir(k, img_dir):
                if os.path.isdir(img_dir):
                    img_list = get_img_files(img_dir)
                    if len(img_list) > 0:
                        tmp_yaml[k] = img_dir

            tmp_yaml = dict()

            get_key_dir('train', train_img_dir)
            get_key_dir('val', val_img_dir)

            if not tmp_yaml.get('train', ''):
                tmp_yaml['train'] = tmp_yaml.get('val', '')

            if tmp_yaml.get('train', '') != '':
                get_key_dir('test', test_img_dir)
                tmp_yaml['names'] = self.names
                tmp_yaml['nc'] = self.nc

                yaml_dict, yaml_base_dir = check_yolov5_yaml(tmp_yaml,
                                                             base_dir=self.base_dir)
                if yaml_dict:
                    print(
                        f'Словарь {yaml_dict}, прошел проверку и установлен как основной')
                    self.set_class_param(yaml_dict, yaml_base_dir, yaml_path=None)
                    return True
            return False

        '''                
        if check_tmp_yaml():
            return train_img_dir, val_img_dir, test_img_dir
        '''

        img_dict = explore_img_dir(self.base_dir)
        # print('get_img_dir_list img_dict', img_dict)
        train_img_dir, val_img_dir, test_img_dir = img_dir_name_analyze(img_dict,
                                                                        f_name_prior=True)
        # print('get_img_dir_list img_dict', train_img_dir, val_img_dir, test_img_dir)

        if check_tmp_yaml():
            return train_img_dir, val_img_dir, test_img_dir
        return
        '''
        # если дошли сюда возможен вариант когда аннотации заданы в других каталогах
        # здесь мы проверяем аннотации yolo, другие форматы - пока пропускаем
        for img_dir in [train_img_dir, val_img_dir, test_img_dir]:
            if check_yolov5_txt_for_img(img_dir, img_lab_dir_dict[img_dir]):
                new_lab_dir = get_yolo_lab_dir(img_dir)
                os.makedirs(new_lab_dir, exist_ok=True)
                source_dir = img_lab_dir_dict[img_dir]
                print(f'копируем аннотации из каталога {source_dir} в {new_lab_dir}')
                for f_name in os.listdir(source_dir):
                    copyfile(os.path.join(source_dir, f_name), new_lab_dir)
                continue
            else: return train_img_dir, val_img_dir, test_img_dir

        #возможно удалось скопировать новые метки, проверяем в третий раз
        check_tmp_yaml()
        return train_img_dir, val_img_dir, test_img_dir
        '''

    def save_yaml(self, yaml_path='', yaml_file='', f_overwrite=False):
        if self.f_autosave:
            if not yaml_path:
                yaml_path = self.new_yaml_path if self.new_yaml_path else self.base_dir
            if not yaml_file:
                yaml_file = self.new_yaml_f_name
            f_overwrite = self.f_overwrite
        dest_yaml_path = os.path.join(yaml_path, yaml_file)
        if dest_yaml_path:
            print(f'сохранен новый словарь {dest_yaml_path}')
            save_data_yaml(self.yaml_dict, dest_yaml_path, f_overwrite)

    def set_class_param(self, yaml_dict, base_dir, yaml_path=None):
        # смысл такой, что path может и не быть, а base_dir можно вычислить или задать
        # base_dir - это каталог path из yaml, если такой есть либо каталог
        if isinstance(yaml_path, (dict)):
            self.yaml_path = None
        else:
            self.yaml_path = yaml_path
        self.base_dir = base_dir
        self.yaml_dict = yaml_dict
        self.save_yaml()

    def explore_base_dir_yaml(self):
        # если пришел каталог, то надо  первую очередь искать yaml
        # потому что без него нет информации по классам
        yaml_list = [f_name for f_name in os.listdir(self.base_dir) if
                     os.path.splitext(f_name)[1] == '.yaml']
        if yaml_list:
            print(f' В каталоге {self.base_dir}, найдены файлы описаний: {yaml_list}')
            for yaml_f_name in yaml_list:
                yaml_path = os.path.join(self.base_dir, yaml_f_name)
                yaml_dict, yaml_base_dir = check_yolov5_yaml(yaml_path)
                if yaml_dict:
                    print(
                        f'Словарь {yaml_f_name}, прошел проверку и установлен как основной')
                    self.set_class_param(yaml_dict, yaml_base_dir, yaml_path)
                    return
                else:
                    print(f'Словарь {yaml_f_name}, не прошел проверку')
        # если сюда дошли, значит подходящие yaml не обнаружены, нужно анализировать изображения

    def get_labels(self, train_img_dir, val_img_dir, test_img_dir, format_key):
        # ann_ext_dict = {'yolov5': '.txt', 'coco': '.json', 'pascal_voc': '.xml'}
        ann_f_path_list = run_fast_scandir(self.base_dir, ann_ext_dict[format_key])
        if format_key == 'coco':
            # в одном файле много меток, берем список файлов
            ann_f_path_list = ann_f_path_list[1]
        elif format_key == 'pascal_voc':
            # в одном каталоге много меток, берем список каталогов
            ann_f_path_list = ann_f_path_list[0]
        # print(format_key, ann_f_path_list)
        tot_class_dict, tot_box_dict = dict(), dict()
        for ann_path in ann_f_path_list:
            # class_dict, img_dict, box_dict
            class_dict, _, box_dict = get_yolov5_ann(ann_path, input_format=format_key)
            tot_class_dict.update(class_dict)
            tot_box_dict.update(box_dict)

        # print(tot_class_dict)
        # print(len(tot_box_dict))
        tmp_dict = dict()
        if len(tot_class_dict) < 1:
            print(f'В {ann_f_path_list} не найдены описания классов')
            return False
        else:
            tmp_dict['names'] = list(tot_class_dict.values())
            tmp_dict['nc'] = len(tmp_dict['names'])

        def save_labels(img_dir, f_save=self.f_autosave):
            img_f_list = get_img_files(img_dir)
            dir_img_boxes = {k: v for k, v in tot_box_dict.items() if k in img_f_list}
            ann_nbr = len(dir_img_boxes)
            if ann_nbr > 0:
                lab_dir = get_yolo_lab_dir(img_dir)
                if f_save:
                    if not os.path.isdir(lab_dir):
                        print(f'каталог {lab_dir} не найден, создаем')
                        os.makedirs(lab_dir)
                    save_yolo_ann(lab_dir, dir_img_boxes)
                else:
                    print(
                        f'Для изображений из каталога {img_dir} найдены, но не сохранены {ann_nbr} аннотаций. Параметр f_autosave не установлен')

            else:
                print(f'Не найдены аннотации для изображений из каталога {img_dir}')

        for img_dir in set([train_img_dir, val_img_dir, test_img_dir]):
            if img_dir:
                save_labels(img_dir)

        tmp_dict['train'] = train_img_dir + os.sep
        tmp_dict['val'] = val_img_dir + os.sep
        if test_img_dir: tmp_dict['test'] = test_img_dir + os.sep
        tmp_dict['path'] = self.base_dir

        yaml_dict, _ = check_yolov5_yaml(tmp_dict, base_dir=self.base_dir)
        if yaml_dict:
            self.set_class_param(yaml_dict, self.base_dir, yaml_path=None)
            return True

        return False

# yolo_v5_dataset = YoloV5Dataset(yaml_path_dir=r'C:/WORK/NU/YOLO/data/10K/')
# yolo_v5_dataset = YoloV5Dataset(yaml_path_dir=r'C:/WORK/NU/YOLO/data/10K_no_yaml/')
# yolo_v5_dataset = YoloV5Dataset(yaml_path_dir=r'C:/WORK/NU/YOLO/data/10K_no_yaml/', names=['варон', 'лисиц'])
# yolo_v5_dataset = YoloV5Dataset(yaml_path_dir=r'C:/WORK/NU/YOLO/data/10K_no_yaml/', names=['варон', 'лисиц'])
# yolo_v5_dataset = YoloV5Dataset(yaml_path_dir=r'C:/WORK/NU/YOLO/data/BCCD_coco/')#, names=['варон', 'лисиц'])
# yolo_v5_dataset = YoloV5Dataset(yaml_path_dir=r'C:/WORK/NU/YOLO/data/BCCD_images/')


import os
import random
import shutil
from math import ceil

import matplotlib.pyplot as plt
import matplotlib.patches as patches

import numpy as np
from PIL import Image

#print(f'{__file__}, {__name__}, {__package__}')
""" Набор функций для манипулирования  датасетами yolov5"""

def get_lab_classes(lab_list):

    #Функция принимает список меток yolov5 и возвращает список первых значений как список классов

    return [int(line.split(' ')[0]) for line in lab_list]


def get_file_labels(lab_path):
    """
    Функция парсит координаты из текстового файла

    :param lab_path: путь к файлу аннотаций
    :return:
    lab_list:list - список списков, где каждый элемент списка соответствует метке,
    первое значение  каждого вложенного списка - классу объекта
    следующие значения  - координатам рамок
    lab_classes_list: - список классов объектов приведенных к типу int,
    соответствующих каждой рамке

    NOTE: метки не проверяются, в результат попадут списки соответствующие всем строкам
    NOTE: классы и координаты в lab_list типа строка, в классы в lab_classes_list - int
    """

    lab_list, lab_classes_list = [], []
    if os.path.isfile(lab_path):
        with open(lab_path, 'r') as ann_file:
            lab_list = ann_file.read()
            lab_list = lab_list.splitlines()
            lab_list = [lab for lab in lab_list]
            lab_classes_list = get_lab_classes(lab_list)
    return lab_list, lab_classes_list


def get_filtered_label(lab_path, class_filter: list):
    """

    :param lab_path: путь к файлу с аннотацией
    :param class_filter: массив вида [class_id_1_int, class_id_2_int, ...]
    :return: список списков где элементы вложенного списка (типа str) соответствуют
    координатам bbox. Первый - классу объекта, остальные - координатам
    NOTE: координаты  НЕ проверяются, в результат попадут списки соответствующие
    всем строкам, удовлетворяющим фильтру
    """
    res = []
    if os.path.isfile(lab_path):
        lab_list, lab_classes_list = get_file_labels(lab_path)
        mask = np.in1d(lab_classes_list, class_filter)
        if np.any(mask) == True:
            res = list(np.array(lab_list)[mask])
    return res




def xywhn2xyxy_uno(x, w=640, h=640):

    """
    Преобразование координат bbox для вывода рамок объекта
    переработано из утилит yolov5
    # from general.py import xywhn2xyxy
    в yolo все функции выполнены под матричные и тензорные операции

    # Convert nx4 boxes from [x, y, w, h] normalized to [x1, y1, x2, y2] where xy1=top-left, xy2=bottom-right
    # y = x.clone() if isinstance(x, torch.Tensor) else np.copy(x)

    в тч для наглядности переделано под едничные значения _uno
    """


    y = np.copy(x)
    y[0] = w * (x[0] - x[2] / 2)  # + padw  # top left x
    y[1] = h * (x[1] - x[3] / 2)  # + padh  # top left y
    y[2] = w * (x[0] + x[2] / 2)  # + padw  # bottom right x
    y[3] = h * (x[1] + x[3] / 2)  # + padh  # bottom right y
    return y


def xywh2xywhn_uno(x, w=640, h=640):
    # def get_yolov5_box(img_size, box):

    # COCO "bbox" : [x,y,width,height]
    # x = (x_min, y_min, wth, ht) координаты левого верхнего угла рамки в пикселях

    # img_wth, img_ht = img_size
    # x_min, y_min, wth, ht = box

    y = x.copy()
    y[2] = y[2] / w  # меняем на относительные значения ширину
    y[3] = y[3] / h  # и высоту рамки
    y[0] = y[0] / w + y[2] / 2  # и рассчитываем координаты центра рамки
    y[1] = y[1] / h + y[3] / 2

    return y


def transform_label(trans_dict, label_dir, new_label_dir=None, class_filter=None):

    """
    Экспериментальная заготовка модуля трансформации координат


    :param trans_dict:
    :param label_dir:
    :param new_label_dir:
    :param class_filter:
    :return:
    """
    if not (new_label_dir):
        new_label_dir = label_dir
        #print(new_label_dir)

    if os.path.isdir(new_label_dir) == False:
        os.mkdir(new_label_dir)

    # xywhn метки ожидаются
    # trans_dict.keys = 'stretch'
    def stretch(class_xywh_list, param):
        """
        Растяжение (сжатие) рамки, умножает предпоследнюю и последнюю координату рамки
        на множитель param
        в перспективе множители могли бы быть разные по ширине и высоте
        а также адаптивные к размеру метки и изображения

        :param class_xywh_list:
        :param param:
        :return:
        """
        # растягиваем рамку на заданый  множитель param
        # print(params)
        return class_xywh_list[:-2] + [str(float(class_xywh_list[-2]) * param),
                                       str(float(class_xywh_list[-1]) * param)]

    metaop_dict = dict()
    metaop_dict['stretch'] = stretch

    for k, param in trans_dict.items():
        for f_name in os.listdir(label_dir):
            lab_list, _ = get_file_labels(os.path.join(label_dir, f_name))
            new_lab_list = []
            for label in lab_list:
                if class_filter:
                    # print(f_name, lab_list, class_filter)
                    # if (class_filter): print('class_filter')
                    # break
                    ann_list = label.split()
                    if int(ann_list[0]) in class_filter:
                        new_ann = metaop_dict[k](ann_list, param)
                        new_lab_list.append(' '.join(new_ann))
                else:
                    new_lab_list.append(label)
            if len(new_lab_list) > 0:
                #отладочно if len(new_lab_list) > 1:
                    #print(f_name, len(new_lab_list))
                with open(os.path.join(new_label_dir, f_name), 'w') as f:
                    # print(new_lab_list)
                    f.write('\n'.join(new_lab_list))


def copy_labels(source_lab_dir, dest_lab_dir,
                class_filter=None,
                filter_label=True
                ):
    """
    Функция копирует файлы с аннотациями, содержащими классы из class_filter
    из source_lab_dir в dest_lab_dir

    :param source_lab_dir:
    :param dest_lab_dir:
    :param class_filter:
    :param filter_label: если True - из файлов аннотаций будут также удалены метки
    классов, не входящих в class_filter
    :return:

    NOTE: если class_filter не задан будут скопированы все файлы

    """
    if (os.path.isdir(source_lab_dir) == False) or (
            len(os.listdir(source_lab_dir)) == 0):
        print(source_lab_dir, 'не найден или пуст каталог с метками')
        return

    if os.path.isdir(dest_lab_dir) == False:
        # будет ошибка если нет каталого верхнего уровня
        os.mkdir(dest_lab_dir)

    if class_filter == None:
        # тут и выше нужен фильтр по txt
        #shutil.copytree(source_lab_dir, dest_lab_dir)
        for f_name in os.listdir(source_lab_dir):
            if os.path.isfile(os.path.join(source_lab_dir, f_name)):
                shutil.copy(os.path.join(source_lab_dir, f_name),
                            os.path.join(dest_lab_dir, f_name))
    else:
        for f_name in os.listdir(source_lab_dir):
            source_path = os.path.join(source_lab_dir, f_name)
            lab_filtered_list = get_filtered_label(source_path, class_filter)
            if len(lab_filtered_list) > 0:
                '''
                for lab in lab_filtered_list:
                    if len(lab.split())<5: 
                        print('source_path', source_path, len() )
                        break
                '''
                # !cat {source_path}
                # print('\n**************')
                # print(f_name, lab_filtered_list)
                if filter_label == True:
                    # print(dest_lab_path+os.sep+f_name, lab_filtered_list)
                    with open(os.path.join(dest_lab_dir, f_name), 'w') as lab_filtered_f:
                        # print('\n'.join(lab_filtered_list))
                        # lab_filtered_f.write('\n'.join(lab_filtered_list))
                        lab_filtered_f.write('\n'.join(lab_filtered_list))
                else:
                    shutil.copyfile(source_path, dest_lab_dir+f_name)
                # break
            pass


def show_images(img_list, box_list,
                ncols=4,
                show_box=False,
                show_label=0,
                name_list=(),
                img_size=(4, 3),
                show_conf=False,
                round_conf=3
                ):
    """
    Функция выводит изображения из списка img_list с рамками из box_list
    :param img_list: список путей к файлам изображений
    :param box_list: список меток обектов, соответствующих файлам изображений
    :param ncols:
    :param show_box: если True - вывести bbox
    :param show_label: 0 - не выводить метки, 1- выводить номер класса, 2 - имя класса
    :param img_size:
    :return:
    """

    if show_label == 2:
        if len(name_list) < 1:
            print(f"Не заданы имена классов, выводим номера")
            show_label = 1
    nrows = ceil(len(img_list) / ncols)
    # print(img_list, box_list, nrows, ncols)
    figsize = np.array([img_size[0] * ncols, img_size[1] * nrows])
    fig, axs = plt.subplots(nrows, ncols=ncols, figsize=figsize)
    axs = axs.flat
    for ind, f_name in enumerate(img_list):
        ax = axs[ind]
        # print(f_name)
        with Image.open(f_name) as img:
            ax.imshow(img)
            if show_box == True:
                for box in box_list[ind]:
                    box_data = box.split(' ')
                    cl_id = int(box_data[0])
                    box_xywhn = [float(coord) for coord in box_data[1:]]
                    box_xyxy = xywhn2xyxy_uno(box_xywhn,
                                              w=img.width,
                                              h=img.height)
                    boxcolor=[color/255 for color in yolo_colors(cl_id)]
                    ax.add_patch(
                        patches.Rectangle((box_xyxy[0:2]), box_xyxy[2] - box_xyxy[0],
                                          box_xyxy[3] - box_xyxy[1],
                                          linewidth=1,
                                          edgecolor=boxcolor,
                                          facecolor='none'))

                    box_txt=''
                    if show_label > 0:
                        # cl_id = int(box_data[0]) и наоборот
                        if show_label < 2:
                            box_txt = box_data[0]
                        else:
                            if cl_id > len(name_list):
                                box_txt = "UNKNOWN LABEL " + box_data[0]
                            else:
                                box_txt = name_list[cl_id]
                        # left_bottom = (box_xyxy[0], box_xyxy[3])
                    if show_conf:
                        try:
                            box_txt += f' conf:{round(float(box_data[5]),round_conf)}'
                        except:
                            print(f'Для {f_name} порог conf нет данных')
                    if box_txt:
                        ax.text(box_xyxy[0]+1, box_xyxy[3]-2, box_txt, color=boxcolor)
    plt.show()
    return


def count_classes_in_lab_dir(lab_dir, show=False, figsize=(10,5), names=None):
    # заодно посмотрим баланс классов

    """
    Подсчет классов объектов в файлах аннотаций из каталога lab_dir
    :param lab_dir: каталог с описаниями
    :param show: флаг диаграммы
    :param figsize: размер диаграммы, если не задан, то по умолчанию
    :param names: имена классов если не задан  - то номера классов
    :return:
    """
    print('lab_dir', lab_dir)
    if os.path.isdir(lab_dir) != True:
        print('Не найден каталог с изображениями')
        return

    box_list = []
    for f_name in os.listdir(lab_dir):
        with open(os.path.join(lab_dir, f_name), 'r') as f:
            boxes = f.readlines()
            try:
                box_list += [int(box.split(' ')[0]) for box in boxes]
            except:
                print(f'{os.path.join(lab_dir, f_name)} не найдены {boxes}')
                return
    value, counts = np.unique(box_list, return_counts=True)

    if show:
        if figsize:
            plt.figure(figsize=figsize)
        if not names:
            names = value
        plt.bar(names, counts)

    return value, counts, len(os.listdir(lab_dir)), lab_dir  # , counts/np.sum(counts)

def cut_labels(lab_dir):
    """
    Проверяет все файлы из каталога lab_dir и оставляет в каждой строке не больше
    пяти значений (четырех пробелов): подразумевается, что в каждой строке останется
    class + xywhn

    :param lab_dir:
    :return:
    NOTE: проверка bbox не производится, например bbox с недостаточным количеством
    координат останутся в файле
    функция появилась под конкретную разметку, скорей всего не нужна
    """

    for f_name in os.listdir(lab_dir):
        with open(os.path.join(lab_dir, f_name), 'r') as f:
            box_list = f.read().splitlines()
            lab_cut = [' '.join(box.split(' ')[:5]) for box in box_list]
        with open(os.path.join(lab_dir, f_name), 'w') as f:
            f.write('\n'.join(lab_cut))
    return


def check_lab_by_name(new_lab_dir, exist_lab_dir):
    """
    Функция по именам файлов без расширений проверяет наличие изображения для
    соответствующего файла аннотации

    :param new_lab_dir: путь к каталогу новых аннотаций предполагается, что эти метки
     будут добавляться в существующий каталог)
    :param exist_lab_dir: путь к каталогу обновляемых меток
    :return: возвращает мнодество общих имкен файлов - если пустое - старые аннотации
    однозначно не будут изменены при добавлении
    """

    # можно также под список директорий любой длины сделать только не ясно нужно ли
    new_lab_f_name_set = set([name for name in os.listdir(new_lab_dir)
                              if name.endswith('txt')])
    exist_lab_f_name_set = set([name for name in os.listdir(exist_lab_dir)
                                if name.endswith('txt')])

    return new_lab_f_name_set & exist_lab_f_name_set


def check_img_for_lab(lab_dir, img_dir):
    """
    Функция по именам файлов без расширений проверяет наличие изображения для
    соответствующего файла аннотации
    """

    lab_f_name_set = set([name[:-4] for name in os.listdir(lab_dir)
                          if name.endswith('txt')])
    img_f_name_set = set([name[:-4] for name in os.listdir(img_dir)
                          if name.endswith('jpg')])
    return lab_f_name_set & img_f_name_set


def add_img_by_lab(lab_dir, img_source_dir, img_dest_dir):
    lab_f_name_set = set([name[:-4] for name in os.listdir(lab_dir)
                          if name.endswith('txt')])

    # можно сделать словарем имя без расширения: полное имя
    img_f_name_set = set([name[:-4] for name in os.listdir(img_source_dir)
                          if name.endswith('jpg')])
    img_f_name_set = img_f_name_set & lab_f_name_set
    print(f'найдено {len(img_f_name_set)} для {len(lab_f_name_set)} файлов аннотаций')

    for f_name in img_f_name_set:
        # файлы *.JPG *.JPg и тп не скопирует
        shutil.copy(os.path.join(img_source_dir, f_name, '.jpg') , img_dest_dir)


def show_class_images(img_dir,
                      show_box=False,
                      lab_dir=None,
                      show_label=0,
                      name_list=(),
                      img_size=(4, 3),
                      class_filter=None, #если не задать - ограничит без сообщения
                      ncols=4,
                      # subdir_list=['train'],
                      img_nbr=8,
                      n_part=1,  # число частей, на которые разобъем данные
                      # номер части питон (0..n_parts-1) если n_parts определено а
                      # parts_nbr - нет: покажем случайную, parts_nbr>n_parts - последнюю
                      part_nbr=None,
                      show_conf=False
                      # не сделал, задумано как интерактивный вывод всех изображений
                      # interact=False
                      ):
    """
    Функция выводит на экран изображения в несколько колонок
    с фильтром по классам class_filter
    :param img_dir: путь к изображениям
    :param show_box: флаг вывода bbox
    :param lab_dir: путь к меткам, нужен независимо от значения show_box
    :param show_label: способ вывода меток - 0 - не показывать, 1 - номер, 2 -имя класса
        (из name_list)
    :param name_list: имена классов, если не заполнены, то будут номера
    :param img_size: размер одного изображения
    :param class_filter: фильтр для изображений, если не задан то будет отфильтрован (0)
    :param ncols: число колонок для вывода
    :param img_nbr: число изображений для вывода, если  img_nbr==0 - выведутся все
        изображения, подходящие фильтру классов
    :param n_part: если не задан параметр img_nbr, то весь список разобъется на
    n_part частей
    :param part_nbr: номер части для вывода на экран из списка, разделенного на n_part
    частей
    :return:
    NOTE: параметры n_part и part_nbr применяются для последовательного вывода длинных
    списков изображений по частям

    """

    if class_filter == None:
        print(f'class_filter не задан, устанавливаем class_filter=[0]')
        class_filter = [0]

    print(f"""
            Пробуем вывести {str(img_nbr)+' изображений' if img_nbr>0 
                                        else 'все изображения'}  из каталога {img_dir}, 
            {'с рамками ' if show_box else 'без рамок'},
            с фильтром классов {class_filter}, 
        """)
    if os.path.isdir(img_dir) != True:
        print('Не найден каталог с изображениями')
        return

    if show_box == True:
        #print('lab_dir', lab_dir)
        #print(os.path.isdir(lab_dir))
        if os.path.isdir(lab_dir) != True:
            # Если не задан попытаемся найти относительно изображений
            #new_lab_dir = os.path.abspath(img_dir + os.sep + '..' + os.sep + 'labels'
                                          #+ os.sep)
            new_lab_dir = os.path.join(img_dir, '..', 'labels')
            if lab_dir:
                print(
                    f'Не найден каталог {lab_dir}, пробуем найти метки в каталоге'
                    f' {new_lab_dir}')
            if os.path.isdir(new_lab_dir) == False:
                print('каталог с метками не найден')
                show_box = False
                return
            else:
                lab_dir = new_lab_dir

    if not os.path.isdir(lab_dir):
        print('Не найден каталог с аннотациями - нет данных для фильтра класса')
        return


    #class_filter_arr = np.array(class_filter)
    # def get_other_path():
    img_filtered, box_filtered = [], []
    # for subdir in subdir_list:

    # попробуем набрать в цикле img_nbr картинок для вывода
    for f_name in os.listdir(lab_dir):
        # пока пропускаем случай, когда img_nbr=-1
        with open(os.path.join(lab_dir, f_name), 'r') as ann_file:
            lab_list = ann_file.read().splitlines()
            lab_classes_list = np.array(get_lab_classes(lab_list))
            mask = np.in1d(lab_classes_list, class_filter)
            if np.any(mask) == True:
                # print('len(mask)',len(mask), mask, f_name)
                #img_filtered.append(img_dir + f_name[:-3] + 'jpg')
                img_filtered.append(os.path.join(img_dir, f_name[:-3] + 'jpg'))
                box_filtered.append(list(np.array(lab_list)[mask]))
        # print(img_nbr==False, f_name, len(img_filtered), box_filtered)
        if img_nbr > 0:
            if (len(img_filtered) > 0) & (len(img_filtered) % img_nbr == 0):
                show_images(img_filtered, box_filtered,
                            ncols=ncols,
                            show_box=show_box,
                            show_label=show_label,
                            name_list=name_list,
                            img_size=img_size,
                            show_conf=show_conf
                            )
                img_filtered, box_filtered = [], []
                # набрали img_nbr картинок, вывели их, обнулили списки и прервались
                break

    # если img_nbr==0 тогда нашли все картинки и выбираем только часть из них
    # или если img_nbr>0 но нужного количества rfhnbyjr не нашли - выведем, те, что нашли

    if len(img_filtered) > 0:
        print(len(img_filtered), ' images filtered from classes ', class_filter)
        start_idx, end_idx = 0, None  # покажем все если что-то пойдет не так
        if img_nbr <= 0:
            #print(img_nbr, n_part)
            if n_part > 0:
                part_len = len(img_filtered) / n_part
                if (part_nbr == None) or (part_nbr < 0):  # если n_part задано, а part_nbr не задано или меньше 0 - покажем случайную часть из n_part
                    part_nbr = random.randint(0, n_part - 1)
                # если part_nbr>n_part - покажем последнюю часть из n_part
                elif part_nbr >= n_part:
                    part_nbr = n_part - 1

                start_idx = round(part_len * part_nbr)
                end_idx = round(start_idx + part_len)
        print(f'show images from {start_idx} to {end_idx}')
        show_images(img_filtered[start_idx:end_idx],
                    box_filtered[start_idx:end_idx],
                    ncols=ncols,
                    show_box=show_box,
                    show_label=show_label,
                    name_list=name_list,
                    img_size=img_size,
                    show_conf=show_conf
                    )


# из yolov5 репозитория взято
class Colors:
# Ultralytics color palette https://ultralytics.com/
    def __init__(self):
        # hex = matplotlib.colors.TABLEAU_COLORS.values()
        hexs = ('FF3838', 'FF9D97', 'FF701F', 'FFB21D', 'CFD231', '48F90A', '92CC17', '3DDB86', '1A9334', '00D4BB',
                '2C99A8', '00C2FF', '344593', '6473FF', '0018EC', '8438FF', '520085', 'CB38FF', 'FF95C8', 'FF37C7')
        self.palette = [self.hex2rgb(f'#{c}') for c in hexs]
        self.n = len(self.palette)

    def __call__(self, i, bgr=False):
        c = self.palette[int(i) % self.n]
        return (c[2], c[1], c[0]) if bgr else c

    @staticmethod
    def hex2rgb(h):  # rgb order (PIL)
        return tuple(int(h[1 + i:1 + i + 2], 16) for i in (0, 2, 4))

yolo_colors=Colors()

def show_list_images(img_f_path_list, ncols=2, img_size=(5,5)):

    """
    Вывод изображений (только изображений, без меток) на экран
    :param img_f_path_list: путь к изображениям
    :param ncols: количество колонок дл вывода
    :param img_size: размер одного изображения
    :return:
    """

    nrows = ceil(len(img_f_path_list) / ncols)
    figsize = np.array([img_size[0] * ncols, img_size[1] * nrows])
    fig, axs = plt.subplots(nrows, ncols=ncols, figsize=figsize)
    if len(img_f_path_list) > 1:
        axs = axs.flat
        for ind, f_name in enumerate(img_f_path_list):
            ax = axs[ind]
            with Image.open(f_name) as img:
                ax.imshow(img)
    else:
        with Image.open(img_f_path_list[0]) as img:
            axs.imshow(img)

    plt.show()


def get_img_sizes(img_path, sample_size=0):
    """
    Функция собирает информацию о размерах изображений
    :param img_path: путь к файлам с изображениями
    :param sample_size: количество, если >1 - то количество изображений для вывода
     если sample_size от 0 до 1 - доля от общей выборки
     sample_size = 1.  приравнивается к 1, если нужна вся выборка - можно указать 0
    :return: возвращает массив numpy с именами файлов и двумерный массив с размерами
    """

    if not os.path.isdir(img_path):
        print(f'Путь к файлам изображений не найден: {img_path}!')
        return
    img_f_list = sorted([f_name for f_name in os.listdir(img_path)
                        if os.path.splitext(f_name)[1].lower() == '.jpg'])
    if not len(img_f_list)>0:
        print(f'По адресу {img_path} файлы изображений не найдены!')
        return

    if sample_size >= 1:
        img_f_list = random.sample(img_f_list, min(round(sample_size), len(img_f_list)))
    elif sample_size > 0:
        img_f_list = random.sample(img_f_list, round(sample_size*len(img_f_list)))

    img_wth_list = []
    img_ht_list = []
    for img_f_name in img_f_list:
        img = Image.open(os.path.join(img_path,img_f_name))
        img_wth_list.append(int(img.width))
        img_ht_list.append(int(img.height))
    res = np.array([img_wth_list, img_ht_list]).T

    return res, np.array(img_f_list)

# TODO убрать отладочные сообщения



import os
import random
import shutil
from pathlib import Path

import yaml

from .dataset_utils import count_classes_in_lab_dir, show_list_images, \
    show_class_images
from os_com import os_com


# print(f'{__file__}, {__name__}, {__package__}')

def get_last_exp_dir(proj_path, name='exp'):
    """
    :param proj_path: путь для поиска - соответствует пути для параметра --project
    :param name: соответствует параметру --name
    :return: путь к каталогу с последним экспериментом
    NOTE: установка и сброс параметра --exist-ok, сбои и прерывания работы программы,
    редактирование содержимого диска могут приводить к неправильному результату
    следующая версия этой функции должна искать веса и предлагать последнюю по времени
    сохранения папку weight

    # ниже как имя подкаталога задается в yolo
    # ищем по имени каталога (по умолчанию - 'exp') каталог с максимальным индексом
    #for n in range(2, 9999):
        #p = f'{path}{sep}{n}{suffix}'  # increment path
        #if not os.path.exists(p):  #
            #break
    #path = Path(p)

    """


    last_exp_dir = ''
    if os.path.isdir(proj_path):
        last_exp_dir = proj_path + os.sep + name
        if os.path.isdir(last_exp_dir):
            # был как минимум один эксперимент
            exp_dir_list = sorted([
                exp_dir for exp_dir in os.listdir(proj_path) if
                (os.path.isdir(proj_path + os.sep + exp_dir)) &
                (exp_dir.find(name) == 0)
            ])
            exp_dir_list = exp_dir_list[1:]  # удалили каталог '/exp/'
            if len(exp_dir_list) >= 1:
                exp_dir_num = max(
                    [int(exp_dir[len(name):]) for exp_dir in exp_dir_list])
                # print('exp_dir_num', exp_dir_num)
                last_exp_dir = last_exp_dir + str(exp_dir_num)

    return last_exp_dir


def get_model_last_exp_dir(model, sub='train'):
    """
    ищет каталог с последним экспериментом (тип )
    :param model: объект типа TerraYoloV5
    :param sub:  тип эксперимента 'train', 'val', 'test'
    :return: возвращает найденный путь или None
    Note при сложной схеме экспериментов, например, при использовании параметра
    "exist-ok" можно получить ошибочный результат
    """
    #
    exp_dict = model.exp_dict[sub]
    try:
        # в приоритете текущая история модели
        last_best_weights_path = exp_dict['last_exp_path']
    except:
        proj_path = model.model_dir + os.sep + exp_dict.get('def_dir')
        # здесь можно попытаться извлечь name из exp_dict['yolo_param_dict'] но
        # пока не представляю ситуации, что last_exp_path нет,
        # а exp_dict['yolo_param_dict']['name'] - есть
        last_best_weights_path = get_last_exp_dir(proj_path=proj_path, name='exp')
    return last_best_weights_path


def uniquify(path):
    filename, extension = os.path.splitext(path)
    counter = 1

    while os.path.exists(path):
        path = filename + "(" + str(counter) + ")" + extension
        counter += 1

    return path


def save_data_yaml(yaml_dict, dest_yaml_path, f_overwrite=False,
                   key_set=frozenset(['train', 'val', 'nc', 'names'])):
    """
    Заполняет описание датасета по шаблону и сохраняет по указанному пути.

    :param yaml_dict: параметры описания
    :param dest_yaml_path: путь к описанию
    :return:
    """
    '''
    yaml_content = f"""
        train: {yaml_dict['train']}
        val: {yaml_dict['val']}
        test: {yaml_dict.get('test', '')}

        # number of classes
        nc: {yaml_dict['nc']}

        # class names
        names: {yaml_dict['names']}
        """
    '''
    keys_missing = key_set - set(yaml_dict.keys())
    if keys_missing:
        print(f'Отсутствуют необходимые ключи словаря {keys_missing}')

    if os.path.exists(dest_yaml_path):
        print(f'файл {dest_yaml_path} существует')
        if not f_overwrite:
            dest_yaml_path = uniquify(dest_yaml_path)

    with open(dest_yaml_path, 'w') as f:
        #f.write(yaml_content)
        yaml.dump(yaml_dict, f, default_flow_style=None)
    print(f'Файл {dest_yaml_path} сохранен')

    return dest_yaml_path



def get_data_dict(yaml_path_or_dict):
    """
    Загружает в словарь файл типа yaml. Применяется для описаний датасетов
    :param yaml_path:
    :return:
    """
    if isinstance(yaml_path_or_dict, (dict)):
        return yaml_path_or_dict
    try:
        with open(yaml_path_or_dict, 'r') as f:
            return yaml.safe_load(f)
    except:
        return {}

def get_lab_img_dir(img_lab_dir):
    """
    Для каталога аканчивающегося на images заменит images на labels
    иначе - заменит ПОСЛЕДНИЙ элемент на images
    :param img_lab_dir: каталог images или  labels
    :return:
    NOTE если не images - проверка на labels не делается
    """
    dir_list = img_lab_dir.split(os.sep)
    if dir_list[-1]=='images':
        return os.sep.join(dir_list[:-1]+['labels'])
    else:
        return os.sep.join(dir_list[:-1]+['images'])

def get_actual_yaml(yaml_path):
    pass


class TerraYoloV5:

    """
    Класс загружает yolov5 и запускает train.py, val.py, detect.py
    скрипты конфинурируются при помощи словарей
    exp_dict['train'], exp_dict['test'], exp_dict['val']

    train_dict=dict()
    train_dict['data'] = data_path #путь к описанию датасета
    train_dict['epochs'] = 50
    #train_dict['img'] = 320

    my_terra_yolov5.run(train_dict, exp_type='train')

    test_dict=dict()
    test_dict['weights'] = os.path.join(
            my_terra_yolov5.exp_dict['train']['last_exp_path'], 'weights', 'best.pt')
    test_dict['source'] = ''#путь
    test_dict['save-txt'] = ''
    my_terra_yolov5.run(test_dict,'test')
    """

    def __init__(self,
                 # data_path: str, #путь к файлу описания датасета.yaml
                 work_dir='.',
                 model_ref='https://github.com/ultralytics/yolov5',
                 log_f_name='model.log',
                 reload=False,
                 backup_dir=None,
                 ):
        """
        Создаем рабочий каталог и загружаем в него модель

        :param work_dir: рабочий каталог
        :param model_ref: ссылка на git
        :param log_f_name: лог - файл, используется для отладки
        :param reload: если True полностью удаляем каталог с моделью и загружаем снова
            если  False перезагружаем только в случае отсутствия каталога с моделью

        """

        if not os.path.exists(work_dir):
            os.makedirs(work_dir, exist_ok=True)
        self.work_dir = os.path.abspath(work_dir + os.sep)

        self.log_f_name = work_dir + os.sep + log_f_name
        self.model_ref = model_ref
        # self.model_name = model_name
        self.reload = reload
        self.model_dir = self.work_dir + os.sep + 'yolov5/'
        self.download()
        self.model_dict = self.get_model_dict()

        # exp_dict = словарь с результатами экспериментов
        self.exp_dict = dict()
        for sub in ['train', 'val', 'test']:
            self.exp_dict[sub] = dict()

        self.exp_dict['train']['def_dir'] = '/runs/train/'
        self.exp_dict['val']['def_dir'] = '/runs/val/'
        self.exp_dict['test']['def_dir'] = '/runs/detect/'

        self.exp_dict['train']['script'] = 'train.py'
        self.exp_dict['val']['script'] = 'val.py'
        self.exp_dict['test']['script'] = 'detect.py'
        self.exp_dict['detect'] = self.exp_dict['test']

        if not backup_dir:
            self.backup_dir = os.path.join(self.work_dir, 'backup')

    def download(self):

        """
        метод в зависимости от значения параметров self.reload
        проверяет наличие каталога self.work_dir и загружает в него модель
        если self.reload == True полностью удаляем каталог с моделью и загружаем снова
        если  self.reload == False перезагружаем только в случае отсутствия каталога
        с моделью
        :return:
        """

        if os.path.isdir(self.model_dir) == True:
            if self.reload == True:
                shutil.rmtree(self.model_dir)
            else:
                # os.chdir(self.model_dir)
                # self.os_com(f'pip install -r requirements.txt')
                # os.chdir(wd)
                return

        wd = os.getcwd()
        os.chdir(self.work_dir)

        self.os_com(f'git clone {self.model_ref}')
        os.chdir(self.model_dir)
        self.os_com('pip install -r requirements.txt')
        os.chdir(wd)
        return

    def get_model_dict(self):

        """

        :return:
        метод возвращает словарь имя файла: путь к найденными файлам описаний моделей
        NOTE: в первом приближении пропустил модели .py
        """

        model_path = Path(self.model_dir + '/models/')
        model_full_path = sorted(list(model_path.rglob('*.yaml')))
        model_names = [str(path)[len(str(model_path)) + 1:].split('.')[0]
                       for path in model_full_path]
        # return model_short_path
        return dict(zip(model_names, model_full_path))

    def get_annotation(self, model_name):

        """
        метод выводит на экран содержимое файла описнаия по имени модели
        :param model_name:
        :return:
        NOTE: пока yaml, словарь если что сделаем
        """

        with open(self.model_dict[model_name], 'r') as f:
            print(f.read())

    def do_experiment(self, command, param_dict, project_dir):

        """

        :param command:
        :param param_dict:
        :param project_dir:
        :return:

        NOTE:
         чуть позже написал для определения каталога с результатами функции
         get_model_last_exp_dir, get_last_exp_dir
         но все равно история экспериментов может влисять на результат
         возможны запуск и работа с одной моделью, прописывание путей в параметрах
         'exist-ok', неудачное завершение экспериментов, потому
         в некоторых сложных случаях последние веса придется искать экспериментатору

        """

        exp_dir_set = set()
        if os.path.isdir(project_dir) == True:
            exp_dir_set = set([dir_name for dir_name in os.listdir(project_dir)
                               if os.path.isdir(project_dir + os.sep + dir_name)])
        # print('exp_dir_set', exp_dir_set)

        self.os_com(command, log_f=False)

        last_train_subdir = ''
        if os.path.isdir(project_dir) == True:

            if 'exist-ok' in param_dict.keys():
                # если параметр 'exist-ok' задан
                # результат эксперимента сохранится в подкаталоге param_dict['name']
                # (по-умолчанию param_dict['name']='exp')
                last_train_subdir = param_dict.get('name', 'exp')
            else:
                # если параметр 'exist-ok' не задан - должен быть один новый каталог
                # но если что-то пойдет не так - не будет ни одного
                now_exp_dir_set = set([
                    dir_name for dir_name in os.listdir(project_dir)
                    if os.path.isdir(project_dir + os.sep + dir_name)])
                exp_dir_list = sorted(list(now_exp_dir_set - exp_dir_set))
                if len(exp_dir_list) > 0:
                    last_train_subdir = exp_dir_list[-1]
        else:
            project_dir = ''

        return project_dir + os.sep + last_train_subdir

    def run(self, param_dict, exp_type='val'):

        """
        Запуск эксперимента
        :param param_dict: словарь параметров эксперимента
        :param exp_type: тип эксперимента, один из ключей self.exp_dict (['train', 'val', 'test'])
        :return:
        """

        arg_str = ''.join(
            [' --' + str(k) + ' ' + str(v) for k, v in param_dict.items()]).strip()
        exp_dict = self.exp_dict[exp_type]

        command = 'python ' + self.model_dir + exp_dict['script'] + ' ' + arg_str
        print('command', command)
        # print(exp_dict)
        project_dir = param_dict.get('project',
                                        self.model_dir + exp_dict['def_dir'])

        exp_dict['last_exp_path'] = self.do_experiment(
                                        command, param_dict, project_dir)
        # сейчас нужно обновить или получить реальный param-dict из json
        exp_dict['user_param_dict'] = param_dict
        exp_dict['yolo_param_dict'] = get_data_dict(
            exp_dict['last_exp_path'] + '/opt.yaml')
        exp_dict['history'] = exp_dict.get('history', []) + [exp_dict]
        return

    def count_labels(self, yaml_path):
        """
        Для всех каталогов датасета из описания вывести подсчет рамок классов
        :param yaml_path:
        :return:
        NOTE: если нужны описания конкретной папки нужно пользоватся
         count_classes_in_lab_dir()
        """
        # может быть сделать декоратор TODO для смены рабочего каталога

        wd = os.getcwd()
        os.chdir(self.work_dir)
        data_dict = get_data_dict(yaml_path)
        for k in self.exp_dict.keys():
            img_path = data_dict.get(k)
            if not img_path is None:
                if os.path.isabs(img_path):
                    lab_path = os.path.join(img_path, '../labels/')
                else:
                    lab_path = os.path.join(self.model_dir, img_path, '../labels/')
                lab_path = os.path.abspath(lab_path)
                if os.path.isdir(lab_path):
                    (classes, label_nbr, file_nbr, lab_path
                     ) = count_classes_in_lab_dir(lab_path)
                    print(
                        f'в каталоге {lab_path} найдены описания: {label_nbr} классов: {classes}')
                else:
                    print(f'каталог {lab_path} не найден')
        os.chdir(wd)

    def show_test_images(self,
                         n_samples=5,
                         img_dir=None,
                         #lab_dir=None,
                         ncols=10,
                         img_size=(3,3)):
        """
        Функция для демонстрации нескольких случайных изображений из каталога img_dir
        :param n_samples: общее количество случайных примеров
        :param img_dir: каталог с изображениями
        #:param lab_dir: каталог с аннотациями, пока не используется
        :param ncols: число колонок в выводе
        :param img_size: размер одного изображения
        :return:
        """

        if img_dir is None:
            img_dir = self.exp_dict['test'].get('last_exp_path', None)

        if os.path.isdir(img_dir):
            file_list = [f_name for f_name in os.listdir(img_dir)
                         if (os.path.isfile(img_dir + os.sep + f_name)) &
                         (f_name[-3:].lower() == 'jpg')]
            img_samples = random.sample(file_list, min(n_samples, len(file_list)))
            ncols = min(n_samples, ncols)
            self.show_val_results(img_dir, img_samples, ncols, img_size)
            '''
            for f_name in img_samples:
                print(img_dir + os.sep + f_name)
                with Image.open(img_dir + os.sep + f_name) as img:
                    plt.imshow(img)
                    plt.show()
            '''
        else:
            print('Каталог с изображениями не найден')
        pass

    def show_val_results(self,
                         img_path=None,
                         img_list=('confusion_matrix.png'),
                         ncols=2,
                         img_size=(10, 10)
                         ):
        """
        Метод выводит изображения из списка img_list из каталога  img_path.
        Задумана для вывода графиков, но можно также пользоваться для вывода результатоы
        по конкретным изображениям, например проблемным


        :param img_path: путь для поиска изображений
        :param img_list: список изображений
        :param ncols: число колонок для вывода изображений
        :param img_size:
        :return:
        """


        if not img_path:
            img_path = self.exp_dict['val'].get('last_exp_path',
                                                    get_model_last_exp_dir(self, 'val'))
        print('img_path', img_path)
        if img_path:
            img_f_path_list = [img_path+os.sep+img_f_name for img_f_name in img_list]
            show_list_images(img_f_path_list, ncols=ncols, img_size=img_size)

    def os_com(self, command: str, log_f=True):
        """
        Запуск команды операционной системы
        :param command: текст команды
        :param log_f: если True - перехват вывода
        :return:
        """
        if log_f == True:
            os.system(command + '>>' + self.log_f_name)
        else:
            #!{command}
            os_com(command)
        return

    def save_last_train_weights(self,
                                backup_dir=None,
                                f_last_name=None,
                                # если задано имя  - копируем, иначе - нет
                                f_best_name=None,
                                # если задано имя  - копируем, иначе - нет
                                f_refresh=False):


        def copy_weight(weight_f_name, new_weight_f_name):
            # если None, то этот файл не нужно сохранять
            if new_weight_f_name != None:
                new_weight_path = os.path.join(backup_dir, new_weight_f_name)
                if not f_refresh:
                    new_weight_path = uniquify(new_weight_path)
                source_f_name = os.path.join(last_weights_path, weight_f_name)
                shutil.copy(source_f_name, new_weight_path)
                print(f'Копируем {source_f_name} в {new_weight_path}')

        if (f_last_name == None) and (f_best_name == None):
            print(f'Не выбраны веса для сохранения')
            return
        last_weights_path = os.path.join(
            get_model_last_exp_dir(self, sub='train'), 'weights')
        if os.path.isdir(last_weights_path):
            if backup_dir == None:
                backup_dir = self.backup_dir
                print(f'каталог для сохранения не задан, сохраняем в каталоге /'
                      f'{backup_dir}')
            if not os.path.isdir(backup_dir):
                print(f'каталог для сохранения {backup_dir} не найден,'
                      f' пробуем создать каталог')
                os.makedirs(backup_dir)
            copy_weight('last.pt', f_last_name)
            copy_weight('best.pt', f_best_name)

    def show_bbox_filtered(self,
                           data_yaml,
                           img_dir=None,
                           lab_dir=None,
                           # show_box=True,
                           show_label=0,
                           img_size=(4, 3),
                           class_filter=None,
                           ncols=4,
                           img_nbr=8,
                           n_part=1,  # число частей, на которые разобъем данные
                           part_nbr=None,
                           show_conf=False
                           ):

        data_dict = get_data_dict(data_yaml)
        if not data_dict:
            print(f'описание данных {data_yaml} не найдено или отсутствует')

        cl_nbr = int(data_dict.get('nc', 0))
        name_list = data_dict.get('names', [])

        if (cl_nbr < 1):
            print(f'Количество классов в описании {data_yaml}={cl_nbr}')
        if len(name_list) != cl_nbr:
            print(
                f'Количество имен классов {len(name_list)} не соответствует числу '
                f'классов {cl_nbr}.')
            if show_label > 1:
                print('Имена классов не могут быть выведены')
                show_label = min(show_label, 1)

        # Нужно обработать class_filter может быть список чисел или список строк
        print('class_filter', class_filter)
        try:
            class_filter = [
                name_list.index(x) if type(x) == str else x for x in
                class_filter]
        except:
            print(f'Невозможно применить фильтр {class_filter}'
                  f' для списка классов {name_list}')

        print('class_filter', class_filter)

        def check_dir(img_dir, lab_dir):
            if not img_dir:
                img_dir = data_dict.get('val')
                print(f'Каталог с изображениями не задан, пробуем {img_dir}')
            if not os.path.isabs(img_dir):
                img_dir = os.path.abspath(os.path.join(self.model_dir, img_dir))
            if os.path.isdir(img_dir):
                if not lab_dir:
                    lab_dir = get_lab_img_dir(img_dir)
                    print(f'Каталог с описаниями не задан, пробуем {lab_dir}')
                if not os.path.isdir(lab_dir):
                    print(f'Каталог с описаниями {lab_dir} не существует')
                    return False
            else:
                print(f'Каталог с изображениями {img_dir} не существует')
                return False
            return img_dir, lab_dir

        res = check_dir(img_dir, lab_dir)
        if res:
            img_dir, lab_dir = res
            show_class_images(img_dir,
                              show_box=True,
                              lab_dir=lab_dir,
                              show_label=show_label,
                              name_list=name_list,
                              img_size=img_size,
                              class_filter=class_filter,
                              ncols=ncols,
                              img_nbr=img_nbr,
                              n_part=n_part,
                              # число частей, на которые разобъем данные
                              part_nbr=part_nbr,
                              show_conf=show_conf
                              )



import os
import json
import numpy as np
import xml.etree.ElementTree as ET
import glob
from pathlib import Path
from .TerraYolo import get_data_dict, save_data_yaml



yolov5_img_formats = ['bmp', 'jpg', 'jpeg', 'png', 'tif', 'tiff', 'dng']
yolov5_vid_formats = ['mov', 'avi', 'mp4', 'mpg', 'mpeg', 'm4v', 'wmv', 'mkv']

format_keys = ['yolov5', 'coco', 'pascal_voc']
yolov5_img_video_formats = ['.' + ext for ext in
                            yolov5_img_formats + yolov5_vid_formats]

sa, sb = f'{os.sep}images{os.sep}', f'{os.sep}labels{os.sep}'  # /images/, /labels/ substrings

ann_templ_dict = {
    'yolov5': '**/*.txt', 'coco': '**/*.json', 'pascal_voc': '**/*.xml'}

ann_ext_dict = {
    'yolov5': '.txt', 'coco': '.json', 'pascal_voc': '.xml'}


def xywh2xywhn_uno(x, w=640, h=640):
    # def get_yolov5_box(img_size, box):

    # COCO "bbox" : [x,y,width,height]
    # x = (x_min, y_min, wth, ht) координаты левого верхнего угла рамки в пикселях

    # img_wth, img_ht = img_size
    # x_min, y_min, wth, ht = box

    y = x.copy()

    y[2] = y[2] / w  # меняем на относительные значения ширину
    y[3] = y[3] / h  # и высоту рамки
    y[0] = y[0] / w + y[2] / 2  # и рассчитываем координаты центра рамки
    y[1] = y[1] / h + y[3] / 2

    return y


def xyxy2xywhn_uno(x, w=640, h=640):
    # VOC PASCAL "bboxes" : [xmin,ymin, xmax, ymax]
    # x = (xmin,ymin, xmax, ymax) координаты левого верхнего и правого нижнего угла рамки в пикселях

    y = x.copy()
    y[2] = (y[2] - y[0]) / w  # относительная ширина
    y[3] = (y[3] - y[1]) / h  # относительная высота рамки
    y[0] = y[0] / w + y[2] / 2  # и рассчитываем координаты центра рамки
    y[1] = y[1] / h + y[3] / 2

    return y


def get_yolo_ann_from_coco(ann_f_path):
    """

    :param ann_f_path: путь к файлу аннотаций coco
    :return:
        class_dict - словарь классов - ключ - номер класса, значение - наименование
        img_dict - словарь параметров изображения, имя файла - ключ,
                значение - список [ширина, высота]
        box_dict - параметры файла, ключ - имя файла, значение - массивы
                [cl, xn, yn, wn, hn, ... может быть что угодно, здесь не провреяется]
    """

    assert os.path.isfile(ann_f_path) == True

    # img_dir = os.path.split(ann_f_path)[0]

    with open(ann_f_path) as cocojson:
        annotations_json = json.load(cocojson)

    class_dict = {cl['id']: cl['name'] for cl in annotations_json['categories']}

    # пока так, возможно понадобится cat['path'] + cat['file_name']
    # print(annotations_json['images'][0].keys(), annotations_json['images'][0])
    img_dict = {img['file_name']: [int(img['width']), int(img['height'])]
                for img in annotations_json['images']}
    # tmp_dict нужен чтоб связать имя файла с id в аннотациях
    tmp_dict = {img['id']: img['file_name'] for img in annotations_json['images']}

    # получаем рамки классов в формате coco
    box_dict = dict([])
    for ann in annotations_json['annotations']:
        if ann.get('bbox'):  # в COCO описании 'bbox' может и не быть
            img_f_name = tmp_dict[ann['image_id']]
            box_dict[img_f_name] = box_dict.get(img_f_name, []) + [
                [ann['category_id']] +
                xywh2xywhn_uno(ann['bbox'], w=img_dict[img_f_name][0],
                               h=img_dict[img_f_name][1])]

    return class_dict, img_dict, box_dict


def get_yolo_ann_from_p_voc(source_path):
    # def vocann2dict(source_path):

    # функция вернет список словарей с именами файлов, размерами и координатами

    # https://blog.paperspace.com/train-yolov5-custom-data/
    # Function to get the data from XML Annotation
    def extract_info_from_xml(xml_file):
        root = ET.parse(xml_file).getroot()

        # Initialise the info dict
        info_dict = {}
        info_dict['bboxes'] = []

        # Parse the XML Tree
        for elem in root:
            # Get the file name
            if elem.tag == "filename":
                info_dict['filename'] = elem.text

            # Get the image size
            elif elem.tag == "size":
                image_size = []
                for subelem in elem:
                    image_size.append(int(subelem.text))

                info_dict['image_size'] = tuple(image_size)

            # Get details of the bounding box
            elif elem.tag == "object":
                bbox = {}
                for subelem in elem:
                    if subelem.tag == "name":
                        bbox["class"] = subelem.text

                    elif subelem.tag == "bndbox":
                        for subsubelem in subelem:
                            bbox[subsubelem.tag] = int(subsubelem.text)
                info_dict['bboxes'].append(bbox)

        return info_dict

    ann_info_dict_list = []
    for f_name in os.listdir(source_path):
        # не проверяются посторонние файлы, будет ошибка
        if os.path.splitext(f_name)[1] != ann_ext_dict['pascal_voc']:
            continue
        f_path = os.path.join(source_path, f_name)
        ann_info_dict_list.append(extract_info_from_xml(f_path))

    img_dict = {img['filename']: img['image_size']
                for img in ann_info_dict_list}  # ['images']

    class_list = []
    box_dict = dict()
    for ann_dict in ann_info_dict_list:
        for box_dict in ann_dict["bboxes"]:
            class_list.append(box_dict['class'])

    class_list = sorted(list(set(class_list)))
    class_dict = {k: v for k, v in enumerate(class_list)}
    class_dict_name_nbr = {v: k for k, v in class_dict.items()}

    box_dict = dict()

    for ann_dict in ann_info_dict_list:
        f_name = ann_dict['filename']
        img_box_list = []
        for tmp_dict in ann_dict["bboxes"]:
            img_box_list.append([class_dict_name_nbr[tmp_dict['class']]] +
                                xyxy2xywhn_uno([tmp_dict['xmin'], tmp_dict['ymin'],
                                                tmp_dict['xmax'], tmp_dict['ymax']],
                                               # width,               height
                                               img_dict[f_name][0], img_dict[f_name][1])
                                )
        box_dict[f_name] = img_box_list

    return class_dict, img_dict, box_dict


# input_formats
get_bbox_from_misc_format = {'coco': get_yolo_ann_from_coco,
                             'pascal_voc': get_yolo_ann_from_p_voc}


def save_yolo_ann(dest_path, box_dict):
    for img_f_name, box_list in box_dict.items():
        lab_f_name = '.'.join(img_f_name.split('.')[:-1]) + '.txt'
        lab_f_path = os.path.join(dest_path, lab_f_name)
        with open(lab_f_path, 'w') as f:
            f.write('\n'.join(
                [' '.join(list(map(str, box))) for box in box_list]))


# def create_yolov5_yaml(class_dict,

def get_yolov5_ann(source_path, input_format='coco'):
    return get_bbox_from_misc_format[input_format](source_path)

def convert_2_yolov5_ann(source_path, dest_path, input_format='coco'):
    class_dict, _, box_dict = get_bbox_from_misc_format[input_format](
        source_path)
    if box_dict:
        if not os.path.isdir(dest_path):
            print(f'каталог {dest_path} не найден, создаем')
            os.makedirs(dest_path)
        save_yolo_ann(dest_path, box_dict)

    return class_dict


'''
def is_coco(ann_dir_file_path):
    #здесь сверяем является ли файл метками coco
    def is_coco_label_correct();
        # а здесь нужно проверить корректны ли сами значения 
        #(наприемр не больше размеров изображения и тп)
        pass

    #Not Implemented
    res = False
    return res


def check_ann_format(ann_file):
    # xyxy xywh xywhn xyxyxn
    pass

def is_yolov5(file_path):
    # тут какой-нибудь pydantic подошел бы
    # нам 

def get_img_json_names(ann_dir):
    img_names = [os.path.splitext(f_name)[0] for f_name in os.listdir(ann_dir)
                        if os.path.splitext(f_name)[1] in yolov5_img_video_formats]
    lab_names = [os.path.splitext(f_name)[0]
                        for f_name in glob.glob(os.path.join(ann_dir, '*.json'))]
    return img_names, lab_names

'''


def is_yolov5_ann(ann_path_list, format_type='detection', f_all_files=False):
    def is_detection(f_path):
        # схему можно расширять, пока проверяем возможность преобразовать типы
        try:
            with open(f_path, 'r') as f:
                labels = np.array(
                    [x.split() for x in f.read().strip().splitlines()],
                    dtype=np.float32)  # labels
            return labels.shape[0]  # число строк
        except:
            # print('ERROR!!!ERROR!!!ERROR!!!ERROR!!!ERROR!!!ERROR!!!')
            return False

    # print(ann_path_list)
    format_check = {'detection': is_detection}
    checker = format_check[format_type]
    res = []

    if f_all_files:
        res = [f_path for f_path in ann_path_list if checker(f_path)]
    else:
        for f_path in ann_path_list:
            if checker(f_path):
                return (f_path)

    return res


def is_coco_ann(ann_path_list, format_type='detection', f_all_files=False):
    def is_coco_data(f_path, coco_key):
        with open(f_path, 'r') as cocojson:
            annotations_json = json.load(cocojson)

        if annotations_json.get('annotations'):
            for img_ann in annotations_json['annotations']:
                if img_ann.get(coco_key):
                    return True
        return False

    coco_keys_dict = {'detection': 'bbox', 'segmentation': 'segmentation'}
    coco_key = coco_keys_dict[format_type]

    res = []
    if f_all_files:
        res = [f_path for f_path in ann_path_list if
               is_coco_data(f_path, coco_key)]
    else:
        for f_path in ann_path_list:
            if is_coco_data(f_path, coco_key):
                return (f_path)
    return res


def is_pascal_voc_ann(ann_path_list, format_type='detection', f_all_files=False):
    def is_detection(f_path):
        xml_root = ET.parse(f_path).getroot()
        # Parse the XML Tree
        for elem in xml_root:
            if elem.tag == "object":
                bbox = []
                for subelem in elem:
                    # if subelem.tag == "name":
                    # bbox["class"] = subelem.text
                    if subelem.tag == "bndbox":
                        for subsubelem in subelem:
                            try:
                                bbox.append(int(subsubelem.text))
                            except:
                                continue
                    if len(bbox) >= 4:
                        return True
        return False

    format_check = {'detection': is_detection}
    checker = format_check[format_type]
    res = []

    if f_all_files:
        res = [f_path for f_path in ann_path_list if checker(f_path)]
    else:
        for f_path in ann_path_list:
            if checker(f_path):
                return (f_path)
    return res


format_ann_check_dict = dict(zip(['yolov5', 'coco', 'pascal_voc'],
                                 [is_yolov5_ann, is_coco_ann, is_pascal_voc_ann]))

def check_ann_format(ann_path_list,
                     format_key, format_type='detection', f_all_files=False):
    # вынес отдельной функцией на случай проверки формата по известным ключам
    return format_ann_check_dict[format_key](ann_path_list,
                                             format_type=format_type,
                                             f_all_files=f_all_files)


def get_ann_format(ann_dir_file_path,
                   format_type='detection', f_all_formats=False):
    """

    :param ann_dir_file_path:
    2 варианта либо каталаг - yolov5, PASCAL_VOC, COCO
    либо файл  COCO, yolov5
    в первом приближении такой алгоритм:
    - если каталог и >=1 txt файла - yolov5
    - если каталог json файлы и соответствующие им изображения - pascal_voc
    - если каталог и >1 json файлов - но неет соответствующих изображений - coco

    в следующих уточнениях можно проверять собственно форматы файлов аннотаций

    - если файл
        - если json - COCO
        - если yaml - yolo + поиск/проверка путей


    :return:
        возвращает 'yolov5' , 'coco', 'pascal_voc' или False
    """

    '''
    def check_yolov5_lab(ann_dir, format_key= f_key, format_type='detection'):

        # работает но очень долго и непонятно ради чего
        #return format_ann_check_dict[f_key](probably_ann_list,
                                            #format_key = f_key,
                                            #format_type='detection')
        return check_ann_format(probably_ann_list, format_key= f_key,
                                                    format_type=format_type)
    '''
    # templ_dict = dict(zip(['yolov5', 'coco', 'pascal_voc'],
    #                            ['**/*.txt', '**/*.json', '**/*.xml']))

    # порядок ключей  имеет значение - сначала нужно проверить что нет описания yolo
    res = []
    if os.path.isdir(ann_dir_file_path):
        for f_key in format_ann_check_dict.keys():
            probably_ann_list = glob.glob(os.path.join(ann_dir_file_path,
                                                       ann_templ_dict[f_key]),
                                          recursive=True)
            # print(probably_ann_list)
            if check_ann_format(probably_ann_list, f_key,
                                format_type=format_type, f_all_files=False):
                if f_all_formats:
                    res.append(f_key)
                    continue
                else:
                    return [f_key]

    elif os.path.isfile(ann_dir_file_path):
        # Not implemented yet
        # пока задача стоит анализировать директорию
        pass
    pass

    return res


# в yolov5 в описании датасета указывается только пути к ИЗОБРАЖЕНИЯМ
# ищем всех родителей и  на уровне каталога images создаем каталог labels
# если images не находим - создаем на уровне родительского
# и, с учетом флага f_replace копируем в этот каталог метки изображений
# тут тоже возможны варианты типа родительский каталог может быть общим для
# нескольких дргуих и тогда смешаются метки разных катлогов в один

# 1. найдем все картинки
# 2. найдем все каталоги картинок
# 3. каждoму каталогк сопоставим каталог меток
# 4. далее по списку - создаем каталог меток - создаем в нем метки соответствующих изображений
# 5. мб все через pandas

# это заготовка для автоматического создания структуры каталогов
# этот код не безопасен так как не проверяются каталоги из списка
# на вложенность, потому переименование одного может поменять имена других
'''
def change_structure(img_dir_list, sa, f_rename=False):#, sb
    for dir_d in img_dir_list:
        # len(dir_d.rsplit(sa, 1)) если >1, то нужная подстрока sa содержится 
        # в пути и каталог dir_d переделывать не нужно, пропускаем
        if len(dir_d.rsplit(sa, 1))==1: 
            # запоминаем содержимое каталога
            # тут есть варианты запоминать ли только картинки или все файлы
            # пока не заморачиваемся и запомнинаем все
            f_name_list=os.listdir(dir_d)

            #new_img_dir = os.path.join(dir_d, sa)
            new_img_dir = dir_d + sa# path.join не работает тк sa ограничена os.sep = пути из корневого каталога
            print(new_img_dir)
            os.makedirs(new_img_dir, exist_ok=True)
            # имя нового каталога нужно исключить из списка копируемых файлов, 
            # на случай, если оно там было
            new_dir_name = os.path.split(new_img_dir)[1]
            if new_dir_name in f_name_list: 
                                            f_name_list.remove()

            print(f'Перемещаем {len(f_name_list)} файлов из {dir_d} в {new_img_dir}')
            for f_name in f_name_list:
                shutil.move(os.path.join(dir_d, f_name), new_img_dir)
        else:
            print(f'не меняем каталог {dir_d}')
        #break
'''


def get_yolov5_images_glob(src_img_path, f_change_dir_structure=False,
                           f_rename=False):
    # брак

    # это заготовка для автоматического создания структуры каталогов
    # не доделано
    '''
    def del_parent_dir(dir_list):
        def has_parent():

        for fir img_path_list
    '''

    img_path_list = [
        f_path for f_path in glob.glob(os.path.join(src_img_path, '**/*'),
                                       recursive=True)
        if os.path.splitext(f_path)[1] in yolov5_img_video_formats]

    img_dir_list = sorted(set([os.path.split(f_path)[0]
                               for f_path in img_path_list]))
    # если каталог кончается на images, os.path.split(f_path)[0] вернет
    # '... /images' на конце, а нужно '... /images/' потому
    img_dir_list = [img_dir + os.sep for img_dir in img_dir_list]
    # print(src_img_path, img_dir_list)
    # заготовка для автоматического создания структуры каталогов не доделано
    '''
    # в зависимости от того, что мы получили на вход
    # пропускаем вложенные каталоги среди найденных, для минимизации 
    # img_dir_list = del_parent_dir(img_dir_list):
    '''

    # то что ниже навязано кодом из yolov5 utils/dataloaders.py
    '''
    def img2label_paths(img_paths):
        # Define label paths as a function of image paths
        sa, sb = f'{os.sep}images{os.sep}', f'{os.sep}labels{os.sep}'  # /images/, /labels/ substrings
        return [sb.join(x.rsplit(sa, 1)).rsplit('.', 1)[0] + '.txt' for x in img_paths]
    '''
    # нужно подготовить структуру каталогов для аннотаций,
    # соответствующую структуре для изображений

    # sa, sb = f'{os.sep}images{os.sep}', f'{os.sep}labels{os.sep}'  # /images/, /labels/ substrings
    # sa, sb = 'nu_yolo', '/labels/'
    yolov5_lab_dir_list = [sb.join(x.rsplit(sa, 1)) for x in img_dir_list]

    # если путь к изображениям == путь к аннотациям - это означает, что нужной
    # подстроки sa ('/images/') не было в имени img_dir и нужно править
    # структуру каталогов,  иначе yolov5 не заведется

    bad_img_dir_idx = [i for i in range(len(img_dir_list))
                       if img_dir_list[i] == yolov5_lab_dir_list[i]]
    if bad_img_dir_idx:
        # дальнейшие варианты очень мудреные, предложим переименовать каталог
        print(f'Пути к изображениям yolov5 должны содержать подстроку {sa}')
        # возможны варианты когда некоторые пути называются правильно
        # но, чтобы не усложнять их рассматривать не будем
        return [], [], [], []

    # Not Implemented - не доделано
    # заготовка для автоматического создания структуры каталогов
    # правка разрешена, если выставлен флаг f_change_dir_structure
    '''
    # вкратце, если понадобится автоматическое переименование, то
    # нужно будет найти все 'плохие'  каталоги
    # проанализировать их на вложенность и или делать для всех каталогов 
    # верхнего уровня них каталог labels (и он может оказаться общим для 
    # нескольких)
    # или поднимась к корню делать отдельный для каждого
    # и среди них некоторые тоже могут стать общими

    for ind, img_dir in enumerate(img_dir_list):
        if img_dir == lab_dir_list[ind]:
            print(f'Путь к каталогам изображений yolov5 не содержит подстроку {sa}')
            if f_change_dir_structure:
                print('Адаптируем структуру каталогов....')
                change_structure(img_dir_list, sa, f_rename=f_rename)
                #change_structure(img_dir_list, 'nu_yolo')

            else:
                lab_dir_list[ind]=''
            return []
    '''

    ann_ext_list = [os.path.splitext(v)[1] for v in ann_templ_dict.values()]
    ann_path_list = [
        f_path for f_path in glob.glob(os.path.join(src_img_path, '**/*'),
                                       recursive=True)
        if os.path.splitext(f_path)[1] in ann_ext_list]

    # каталоги из списка img_dir_list существуют реально
    # каталоги из списка lab_dir_list могут существовать, а могут и нет
    return img_path_list, img_dir_list, yolov5_lab_dir_list, ann_path_list


def create_yolov5_yaml(data_dir,
                       img_path_list, img_dir_list, lab_dir_list,  # probably_ann_list,
                       dest_yaml_path,
                       class_dict,
                       yaml_f_name_path='dataset.yaml',
                       yaml_path_param=None,  # ключ path
                       f_overwrite=False):
    yaml_dict = dict()

    # ищем среди имен каталогов train, test, val(valid)
    # sa = f'{os.sep}images{os.sep}'#  /images/
    train_dict = dict()
    val_dict = dict()
    test_dict = dict()
    img_nbr_dict = dict()

    for img_dir in img_dir_list:
        path_parts_reversed = Path(img_dir).resolve().parts[::-1]
        # если среди каталогов есть 'train'
        if 'train' in path_parts_reversed:
            # добавляем в соотвествующий словарь максимальный уровень
            train_dict[img_dir] = len(path_parts_reversed
                                      ) - path_parts_reversed.index('train')

        # val
        val_pos = -1
        if 'val' in path_parts_reversed:
            val_pos = len(path_parts_reversed
                          ) - path_parts_reversed.index('val')

        if 'valid' in path_parts_reversed:
            # добавляем в соотвествующий словарь максимальный уровень
            val_pos = max(len(path_parts_reversed
                              ) - path_parts_reversed.index('valid'), val_pos)

        if val_pos >= 0:
            val_dict[img_dir] = val_pos
        # test

        if 'test' in path_parts_reversed:
            # добавляем в соотвествующий словарь максимальный уровень
            test_dict[img_dir] = len(path_parts_reversed
                                     ) - path_parts_reversed.index('test')

        img_nbr_dict[img_dir] = len(
            [f_name for f_name in os.listdir(img_dir)
             if os.path.splitext(f_name)[1] in yolov5_img_video_formats]
            # [f_name for f_name in img_path_list
            # if os.path.split(f_name)[0]==img_dir]
        )

    if train_dict:
        yaml_dict['train'] = list(train_dict.keys())[
            list(train_dict.values()).index(max(train_dict.values()))]
        if yaml_dict['train'] in val_dict.keys():
            del val_dict[yaml_dict['train']]
        if yaml_dict['train'] in test_dict.keys():
            del test_dict[yaml_dict['train']]

        # если train не найден, то остальное ('val', 'test') игнорируем
        if val_dict:
            yaml_dict['val'] = list(val_dict.keys())[
                list(val_dict.values()).index(max(val_dict.values()))]
            if yaml_dict['val'] in test_dict.keys():
                del test_dict[yaml_dict['val']]

        if test_dict:
            yaml_dict['test'] = list(test_dict.keys())[
                list(test_dict.values()).index(max(test_dict.values()))]

    # если не находим по названиям - считаем картинки и в train назначаем max
    if not yaml_dict.get('train', []):
        if img_nbr_dict:
            yaml_dict['train'] = list(img_nbr_dict.keys())[
                list(img_nbr_dict.values()).index(max(img_nbr_dict.values()))]
        else:
            print('не найдены каталоги с изображениями')
            return False

    if not yaml_dict.get('val', []):
        if len(img_nbr_dict) > 1:
            sorted_dir_size = sorted(img_nbr_dict.values())
            val_size = sorted_dir_size[-2]  # -1 - макс - оставлен для train
            yaml_dict['val'] = list(img_nbr_dict.keys())[
                list(img_nbr_dict.values()).index(val_size)]
        else:
            yaml_dict['val'] = yaml_dict['train']
            print(
                f'так как найден только один каталог с изображенями в yaml ключ "val" будет равен "train" и равен {yaml_dict["val"]}')

            # если нет катлога с именем test, по размеру не будем его назначать

    # И как только я это все сделал - попал в ситуацию, когда каталог для валидации назывался test
    # короче это все нужно бросить я думаю
    '''        
    if yaml_dict.get('test',[]):
        #test назначим опционально, если есть свободные каталоги
        if len(img_nbr_dict)>2: 
            sorted_dir_size = sorted(img_nbr_dict.values())
            val_size = sorted_dir_size[-3] #-1, -2 - оставлены для train, val
            yaml_dict['val'] = list(img_nbr_dict.keys())[
                                    list(img_nbr_dict.values()).index(val_size)]
    '''
    # если сюда добрались, значит есть yaml_dict['train'] и yaml_dict['val']

    yaml_dict['names'] = list(class_dict.values())
    yaml_dict['nc'] = len(yaml_dict['names'])
    if yaml_path_param: yaml_dict['path'] = yaml_path_param

    # print('yaml_dict', yaml_dict)

    if yaml_f_name_path:
        if os.path.split(yaml_f_name_path)[0]:
            dest_yaml_path = yaml_f_name_path
        else:
            # можно бы проверить, но кажется не такая уж и ценность
            dest_yaml_path = os.path.join(data_dir, yaml_f_name_path)
        save_data_yaml(yaml_dict, dest_yaml_path, f_overwrite=f_overwrite)
    else:
        print('Путь для записи yaml не задан. Файл не сохранен')
        dest_yaml_path = None

    return dest_yaml_path


def create_yolov5_dataset(data_dir,
                          format_key,
                          yaml_f_name='dataset.yaml',
                          f_overwrite=False):
    # брак

    def is_parent(parent, child):
        par_path = Path(os.path.abspath(parent)).resolve().parts[1:]
        ch_path = Path(os.path.abspath(child)).resolve().parts[1:]
        return par_path == ch_path[:len(par_path)]

    loc_templ_dict = dict(zip(['yolov5', 'coco', 'pascal_voc'],
                              ['*.txt', '*.json', '*.xml']))

    print(f'Создаются аннотации в формате yolov5 для каталога {data_dir}')

    img_path_list, img_dir_list, lab_dir_list, ann_path_list = get_yolov5_images_glob(
        data_dir)  # Not Implemented, f_change_dir_structure=False)

    # img_path_list первый список файлов изображений не должен быть пустым и файлы существуют
    # но он нам особо не нужен, дальше только аннотации смотрим
    # img_dir_list второй список каталогов с изображениями не должен быть пустым и каталоги существуют
    # примерно так img_dir_list =  ['/content/images/train', '/content/images/valid'],
    # а вот каталоги третьего списка для меток может и существовать и не существовать
    # примерно так lab_dir_list = ['/content/labels/train', '/content/labels/valid'],

    # сейчас нужно для каждого каталога из img_dir_list найти файл с аннотациями
    # конвертнуть их и сохранить результат в соотвествующую папку из lab_dir_list

    # здесь у нас могут быть два варианта coco или pascal_voc
    # и там и там аннотации содержатся в папке с изображениями

    class_dict_list = []
    for ind, img_dir in enumerate(img_dir_list):
        # так как каталоги предобработаны рекурсивный поиск сейчас не нужен
        '''
        probably_ann_list = glob.glob(os.path.join(img_dir,
                                                #'*.json'),
                                                loc_templ_dict[format_key]),
                                                #ann_templ_dict[f_key]),
                                                recursive=False)

        '''
        probably_ann_list = [f_path for f_path in ann_path_list if
                             os.path.splitext(f_path)[1] == os.path.splitext(
                                 loc_templ_dict[format_key])[1]]
        probably_ann_list = [f_path for f_path in probably_ann_list if
                             is_parent(img_dir, f_path)]

        if probably_ann_list:
            if format_key == 'coco':
                # вообще файлов с аннотациями coco может быть и много, возьмем первый пока
                ann_file = probably_ann_list[0]
                class_dict_list.append(convert_2_yolov5_ann(ann_file,
                                                            lab_dir_list[ind],
                                                            input_format=format_key))
            elif format_key == 'pascal_voc':
                ann_dir_list = sorted(set(
                    [os.path.split(f_path)[0] for f_path in probably_ann_list]))
                # аналогично, возьмем первый
                ann_dir = ann_dir_list[0]
                class_dict_list.append(convert_2_yolov5_ann(ann_dir,
                                                            lab_dir_list[ind],
                                                            input_format=format_key))
        else:
            print(f'В каталоге {img_dir} не найдены аннотации {format_key}')

    # у нас было несколько каталогов, берем самый большой из справочника классов
    # не безопасно, в идеале нужно слить все справочники
    cl_dict_len_list = [len(cl_dict) for cl_dict in class_dict_list if cl_dict]
    if not cl_dict_len_list:
        print(f'В каталогах {img_dir_list} не найдены классы объектов {format_key}')
        return False
    max_cl_dict_len = max(cl_dict_len_list)
    class_dict = class_dict_list[cl_dict_len_list.index(max_cl_dict_len)]

    return create_yolov5_yaml(data_dir,
                              img_path_list, img_dir_list, lab_dir_list,
                              # probably_ann_list,
                              dest_yaml_path=data_dir,
                              class_dict=class_dict,
                              yaml_f_name_path='dataset.yaml',
                              yaml_path_param=None,  # ключ path
                              f_overwrite=f_overwrite)


def check_yolov5_yaml(yaml_path, base_dir=None, force_base_dir=False):
    """
    :param yaml_path: путь к yaml описанию
    :param base_dir: путь для замены переменной path
    :param force_base_dir: если True - переменная path из yaml_path будет проигнорирована
     на случай перемещеия данных
    :return: True
    NOTE: base_dir будет работать только если в параметрах train, val, test
     - неабсолютные пути
     если base_dir == None, вместо base_path будет попытка взять
     os.path.split(yaml_path)[0], если base_dir не нужен - нужно передать ''

    """

    def get_abs_path(dict_dir):
        if not (os.path.isabs(dict_dir)):
            # dict_dir = os.path.join(base_dir, dict_dir)
            dict_dir = os.path.abspath(os.path.join(base_dir, dict_dir))
        return str(dict_dir)  # чтобы прошла проверка путь должен os.sep заканчиваться

    def check_yolo_img_lab_dir(dict_key):
        # img_dir_for_check = get_abs_path(Path(yaml_dict.get(dict_key, '')))
        img_dir_for_check = get_abs_path(yaml_dict.get(dict_key, ''))
        if not os.path.isdir(img_dir_for_check):
            print(f'Каталог {img_dir_for_check} не найден')
            return ''
        img_f_names = set([os.path.splitext(f_name)[0]
                           for f_name in os.listdir(img_dir_for_check)
                           if os.path.splitext(f_name)[1] in yolov5_img_video_formats])
        if len(img_f_names) < 1:
            print(f'В каталоге {img_dir_for_check} не найдены файлы с изображениями')
            return ''

        # чтобы получился путь к меткам, и последний каталог должен быть строго /images/
        # просто /images без слэша в конце будет ошибка. Если не последний - все равно,

        # на случай если последний каталог == /images нужно добавить сепаратор
        img_dir_for_check = str(img_dir_for_check) + os.sep
        lab_dir_for_check = sb.join(img_dir_for_check.rsplit(sa, 1))

        if not os.path.isdir(lab_dir_for_check):
            print(f'Каталог меток {lab_dir_for_check} не найден')
            return ''

        lab_f_names = set([os.path.splitext(f_name)[0]
                           for f_name in os.listdir(lab_dir_for_check)
                           if os.path.splitext(f_name)[1] == '.txt'])

        if len(img_f_names & lab_f_names) < 1:
            print(f'В каталоге {img_dir_for_check} не найдены подходящие аннотации')
            return ''

        return img_dir_for_check

    result = (dict(), None)

    yaml_dict = get_data_dict(yaml_path)
    if not isinstance(yaml_dict, (dict)):
        print(f'По адресу {yaml_path} не найден словарь с описанием')
        return result

    if base_dir == None:
        base_dir = os.path.split(yaml_path)[0]
    if not force_base_dir:
        base_dir = yaml_dict.get('path', base_dir)
    # base_dir = Path(base_dir)

    train_dir = check_yolo_img_lab_dir('train')
    if not train_dir:
        return result
    val_dir = check_yolo_img_lab_dir('val')

    if not val_dir:
        return result

    if val_dir == train_dir:
        print(f'Kлючи "train" и "val" ссылаются на один каталог')

    names = yaml_dict.get('names', [])
    if len(names) < 1:
        print(f'Не найдены наименования классов')
        return result

    nc = int(yaml_dict.get('nc', 0))

    if nc < 1:
        print(f'Не указано количество классов')
        return result
    return yaml_dict, base_dir


def get_actual_yolov5_yaml(yaml_path,
                           dest_yaml_path=None):
    """
    # функция должна проверить yaml в каталоге
    # и не совсем понятно что должна сделать еще, пока оставим

    :param dataset_dir_yaml_path: путь к датасету или yaml описанию
    :param dest_yaml_path: путь для нового описания

    :return: путь к актуальному yaml, старому или созданному
    """




from subprocess import Popen, PIPE, STDOUT
from IPython.core.magic import register_line_magic


@register_line_magic
def colab_os_com(command):

    os.environ["PYTHONUNBUFFERED"] = "1"

    with Popen(
            command, stdout=PIPE, shell=True, stderr=STDOUT, bufsize=1, close_fds=True
    ) as process:
        for line in iter(process.stdout.readline, b""):
            print(line.rstrip().decode("utf-8"))


# old
'''
#import types
def colab_os_com(obj, command, log_f=True):
    if log_f == True:
        os.system(command + '>>' + obj.log_f_name)
    else:
        !{command}
    return
setattr(self, 'os_com', types.MethodType(colab_os_com, TerraYoloV5))
'''
'''
import shlex
from subprocess import STDOUT, check_call

        else:
            # !{command}
            #current_working_dir = os.getcwd()
            #print('current_working_dir', current_working_dir)
            #tmp_str = input()
            #process = subprocess.Popen(shlex.split(command), stdout=subprocess.PIPE)
            #process = subprocess.Popen(command,cwd=current_working_dir, stdout = PIPE, stderr = PIPE, shell = True)#, encoding='utf8'
            #process = subprocess.Popen(command, stdout = PIPE, stderr = PIPE, shell = False)#, encoding='utf8'
            #process = subprocess.Popen(shlex.split(command), cwd=current_working_dir, stdout = subprocess.PIPE, encoding='utf8')#
            #process = subprocess.Popen(shlex.split(command), shell = False, stdout = subprocess.PIPE, encoding='utf8')
            #process = subprocess.Popen(shlex.split(command), stdout=subprocess.PIPE)


            while True:
                output = process.stdout.readline()
                print(output)
                if output == '' and process.poll() is not None:
                    break
                if output:
                    print(output.strip().decode("utf-8"))
            rc = process.poll()
'''


import os
def fast_scandir(dir_name):
    #https://stackoverflow.com/questions/18394147/how-to-do
    # -a-recursive-sub-folder-search-and-return-files-in-a-list/59803793#59803793
    subfolders= [f.path for f in os.scandir(os.path.normpath(dir_name)) if f.is_dir()]
    for dir_name in list(subfolders):
        subfolders.extend(fast_scandir(dir_name))
    return subfolders

def run_fast_scandir(dir_name, ext):    # dir: str, ext: list
    subfolders, files = [], []

    for f in os.scandir(dir_name):
        if f.is_dir():
            subfolders.append(f.path)
        if f.is_file():
            if os.path.splitext(f.name)[1].lower() in ext:
                files.append(f.path)


    for dir_name in list(subfolders):
        sf, f = run_fast_scandir(dir_name, ext)
        subfolders.extend(sf)
        files.extend(f)
    return subfolders, files



from subprocess import Popen, PIPE, STDOUT
from IPython.core.magic import register_line_magic

def non_colab_os_com(command):
    with Popen(
        command, stdout=PIPE, shell=True, stderr=STDOUT, bufsize=1, close_fds=True
    ) as process:
        for line in iter(process.stdout.readline, b""):
            print(line.rstrip().decode("utf-8"))

'''
import types
@register_line_magic
def colab_os_com(obj, command, log_f=True):
    if log_f == True:
        os.system(command + '>>' + obj.log_f_name)
    else:
        colab_os_com(command)
    return
setattr(my_terra_yolov5, 'os_com', types.MethodType(colab_os_com, TerraYoloV5))
'''


#import os

### общие соображения
# класс привязан к структуре каталогов и содержит от одного подраздела подкатлога
# yaml для каталогов имеет вспомогательное значение, для классов - основное
# смысл в том, что класс для того, чтобы получить yaml,
# если yaml есть - то и класс не так чтоб нужен
# ну еще как развитие - хранить результаты в классе, но это не точно
# train - val - test - все остальное misc, в общем случае
# привязываемся к именам - train , val-valid, test
# нет train с аннотациями - нет датасета
# потому если нет имени train - выбираем для train самый большой каталог
# в test может не быть аннотаций
# у val-valid должны быть аннотации, потому, если нет имени val-valid
# но есть test и для test есть аннотации - test становиться val-valid
# если testa нет или нет для него аннотаций - train становится val
#


# sa, sb = f'{os.sep}images{os.sep}', f'{os.sep}labels{os.sep}'# /images/, /labels/ substrings
# from TerraYoloTest.dataset_utils.ann_convertors import check_yolov5_yaml, sa,sb
from TerraYoloTest.dataset_utils.ann_convertors import sa, sb, yolov5_img_video_formats
from TerraYoloTest.TerraYolo import save_data_yaml, get_data_dict


# from TerraYoloTest.utils.general import fast_scandir


def get_yolo_lab_dir(img_dir):
    return sb.join((str(pathlib.Path(img_dir)) + os.sep).rsplit(sa, 1))


class YoloV5Dataset:
    def __init__(self,
                 yaml_path_dir=None,
                 train_img_dir=None, val_img_dir=None, test_img_dir=None,
                 # data_root_path=None,
                 # cl_labels=None,
                 names=None,

                 ):

        if not os.path.exists(yaml_path_dir):
            print(f'файл/каталог {yaml_path_dir} с данными не найден')
            return None

        self.data_keys = ['train', 'val', 'test']
        self.data_dict = dict()
        for k in self.data_keys:
            self.data_dict[k] = dict()

        # на этом шаге ищем каталоги с изображениями
        if os.path.isfile(yaml_path_dir):
            # разница в том, что в parse_yaml сразу проверяется соответствие меток
            self.parse_yaml(yaml_path)
            # те если есть yaml то есть и датасает yolo
        elif os.path.isdir(yaml_path_dir):
            # а если yaml нет, то возможно yolo метки нужно будет еще только сделать
            # if names:
            # else print()
            self.explore_base_dir(base_dir=yaml_path_dir)
            # else:
            # self.explore_dir_list(train_img_dir, val_img_dir, test_img_dir)

        if self.data_dict:
            self.train_dict = self.data_dict['train']
            self.val_dict = self.data_dict.get('val', dict())
            self.test_dict = self.data_dict.get('test', dict())
        else:
            print(f'В файл/каталог {yaml_path_dir} с данными не найден')
            return None

        # на этом шаге ищем каталоги с аннотациями

        # анализ изображений и аннотаций если чего-то не хватает - предложить поискать еще

        # print('self.train_dict', self.train_dict, self.data_dict['train'])
        # yaml_path  в приоритете, если не найдется yaml_dict - пробуем data_root_path
        # else:
        # pass
        # if self.parse_root_path(data_root_path)

        # в этом месте должен быть self.train - из yaml или структуры каталогов
        # нет train - нет датасета

        if data_root_path:
            self.path = data_root_path
        # self.classes
        # self.nc
        pass

    def parse_yaml(self, yaml_path):
        yaml_dict, base_dir = check_yolov5_yaml(yaml_path)
        if yaml_dict:
            # смысл такой, что path может и не быть, а base_dir можно вычислить или задать
            self.base_dir = base_dir
            self.path = yaml_dict.get('path', '')
            # если пройдена check_yolov5_yaml то 'train' и 'val' обязаны быть
            # и соответствующие каталоги с метками тоже и содержать метки
            self.data_dict['train']['img_dir'] = yaml_dict['train']
            self.data_dict['val']['img_dir'] = yaml_dict['val']
            if yaml_dict.get('test'):
                self.data_dict['test']['img_dir'] = yaml_dict['test']

            self.names = yaml_dict['names']
            self.nc = int(yaml_dict['nc'])
            return True
        return False

    def explore_base_dir(self, base_dir, name_priority=True):
        # вообще если пришел каталог, то надо бы в первую очередь искать yaml
        # потому что без него нет информации по классам
        yaml_list = [f_name for f_name in os.listdir(base_dir) if
                     os.path.splitext(f_name) == '.yaml']
        print(yaml_list)

        img_dict = explore_img_dir(base_dir)
        if not img_dict:
            return None

        train_dict, val_dict, test_dict = img_dir_name_analyze(
            img_dir_dict=img_dict)

    def explore_dir_list(self, train_img_dir, val_img_dir, test_img_dir):
        # NotImplemented
        pass

    def get_img_info(self):
        # тут надо загрузить информацию об изображениях и подгружать по мере необходимости новую
        # NotImplemented
        pass

    def train_test_split(self):
        # NotImplemented
        pass

    def save(self, dest_path):
        # сохранить датасет
        # NotImplemented
        pass

    def load(self, source_path):
        # загрузить датасет
        # NotImplemented
        pass

    def save_yaml(self):
        # сохранить yaml
        # NotImplemented
        pass

    def add_exp(self):
        # добавить в датасет результаты эксперимента
        # NotImplemented
        pass


#yolo_v5_dataset = YoloV5Dataset(yaml_path_dir=r'C:/WORK/NU/YOLO/data/10K/')
'''
# yolo_v5_dataset = YoloV5Dataset(yaml_path=r'C:\WORK\NU\YOLO\data\10K\data_path.yaml')
yolo_v5_dataset = YoloV5Dataset(yaml_path=r'C:/WORK/NU/YOLO/data/10K/data_path.yaml')
dir(yolo_v5_dataset)

print('get_yolo_lab_dir', get_yolo_lab_dir(r'C:\WORK\NU\YOLO\data\10K\images\\'))
print(get_yolo_lab_dir(r'C:\WORK\NU\YOLO\data\10K\images'))
print(get_yolo_lab_dir(r'C:\WORK\NU\YOLO\data\images\10K'))
print(get_yolo_lab_dir(r'C:\WORK\NU\YOLO\data\images\10K\\'))
'''



import os
import torch


def torch_load(terra_model,
               repo_or_dir='ultralytics/yolov5',
               sub_model='yolov5s',
               source='github',
               weight_path='',
               pretrained=True, force_reload=False, verbose=True):
    """

    :param repo_or_dir:
    :param sub_model:
    :param source:
    :param weight_path:
    :param pretrained:
    :param force_reload:
    :param verbose:
    :return:
    Note: If ``source`` is 'github', ``repo_or_dir`` is expected to be
    of the form ``repo_owner/repo_name[:ref]`` with an optional
    ref (a tag or a branch).

    If ``source`` is 'local', ``repo_or_dir`` is expected to be a
    path to a local directory. - do not implemented
    """

    wd = os.getcwd()
    os.chdir(terra_model.work_dir)
    if sub_model == 'custom':
        if os.path.isfile(weight_path):
            terra_model.hub_model = torch.hub.load(
                repo_or_dir, sub_model,
                path=weight_path,
                force_reload=force_reload,
                verbose=verbose
            )
            return
        else:
            print(f'веса не {weight_path} найдены, пробуем загрузить')

    terra_model.hub_model = torch.hub.load(
        repo_or_dir, sub_model, source=source,
        pretrained=pretrained,
        force_reload=force_reload,
        verbose=verbose
    )
    os.chdir(wd)
    return
